from datetime import datetime
from enum import Enum
from typing import List, Optional
from uuid import UUID

from pydantic import Field, root_validator

from darwin.future.data_objects.typing import UnknownType
from darwin.future.pydantic_base import DefaultDarwin


class WFDataset(DefaultDarwin):
    """
    A class to manage all the information around a dataset on the darwin platform,
    including validation

    Attributes
    ----------
    id: int
    name: str
    instructions: str

    Methods
    ----------
    __int__: returns the id of the dataset
    __str__: returns the name of the dataset
    __repr__: returns a string representation of the dataset
    """

    id: int
    name: str
    instructions: str = Field(min_length=0)

    def __int__(self) -> int:
        return self.id

    def __str__(self) -> str:
        return self.name


class WFEdge(DefaultDarwin):
    """
    A workflow edge

    Attributes
    ----------

    id: str
    name: str
    source_stage_id: str
    target_stage_id: str
    """

    id: UUID
    name: str
    source_stage_id: Optional[UUID]
    target_stage_id: Optional[UUID]

    @root_validator(pre=True)
    def _one_or_both_must_exist(cls, values: dict) -> dict:
        if not values["source_stage_id"] and not values["target_stage_id"]:
            raise ValueError("One or both of source_stage_id and target_stage_id must be defined")

        return values


class WFType(Enum):
    """
    The type of workflow stage (Enum)

    Enumerations
    ----------
    DATASET: str
    ANNOTATE: str
    REVIEW: str
    COMPLETE: str
    """

    # TODO: There may be more types
    DATASET = "dataset"
    ANNOTATE = "annotate"
    REVIEW = "review"
    COMPLETE = "complete"


class WFUser(DefaultDarwin):
    stage_id: UUID
    user_id: int


class WFStageConfig(DefaultDarwin):
    # ! NB: We may be able to remove many of these attributes
    url: Optional[str]
    x: Optional[int]
    y: Optional[int]

    dataset_id: Optional[int]
    model_type: Optional[str]

    parallel_stage_ids: Optional[List[UUID]]
    readonly: bool

    # Included, and type known, but potentially not involved in backend
    auto_instantiate: bool
    include_annotations: bool
    initial: bool
    retry_if_fails: bool
    rules: List
    skippable: bool
    test_stage_id: Optional[UUID]

    # unsure of needs for these, so they here for future proofing
    allowed_class_ids: UnknownType
    annotation_group_id: UnknownType
    assignable_to: UnknownType
    authorization_header: UnknownType
    champion_stage_id: UnknownType
    class_mapping: List
    from_non_default_v1_template: UnknownType
    iou_thresholds: UnknownType
    model_id: UnknownType
    threshold: UnknownType


class WFStage(DefaultDarwin):
    """
    A workflow stage

    Attributes
    ----------
    id: str
    name: str

    type: WFType

    assignable_users: List[WFUser]
    edges: List[WFEdge]
    """

    id: UUID
    name: str

    type: WFType

    assignable_users: List[WFUser]
    edges: List[WFEdge]


class Workflow(DefaultDarwin):
    """
    A class to manage all the information around a workflow on the darwin platform,
    including validation

    Attributes
    ----------
    id: str
    name: str
    team_id: int

    inserted_at: datetime
    updated_at: datetime

    dataset: WFDataSet
    stages: List[WFStage]

    thumbnails: List[str]
    """

    id: UUID
    name: str
    team_id: int

    inserted_at: datetime
    updated_at: datetime

    dataset: WFDataset
    stages: List[WFStage]

    thumbnails: List[str]


class WorkflowListValidator(DefaultDarwin):
    list: List[Workflow]



from typing import Union

import responses
from pytest import raises
from requests import HTTPError

from darwin.future.core.client import Client
from darwin.future.core.datasets.create_dataset import create_dataset
from darwin.future.data_objects.dataset import Dataset
from darwin.future.exceptions.base import DarwinException
from darwin.future.tests.core.fixtures import *  # noqa: F401, F403

from .fixtures import *  # noqa: F401, F403


def test_it_creates_a_dataset(basic_dataset: Dataset, base_client: Client) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.POST,
            base_client.config.api_endpoint + "datasets",
            json=basic_dataset,
            status=200,
        )

        dataset = create_dataset(base_client, "test-dataset")
        assert dataset.name == "test-dataset"
        assert dataset.slug == "1337"


def test_it_raises_an_error_on_http_error(basic_dataset: Dataset, base_client: Client) -> None:
    with raises(HTTPError):
        with responses.RequestsMock() as rsps:
            rsps.add(
                rsps.POST,
                base_client.config.api_endpoint + "datasets",
                json=basic_dataset,
                status=400,
            )

            create_dataset(base_client, "test-dataset")




from typing import Optional

from darwin.future.core.client import Client
from darwin.future.exceptions.core.datasets import DatasetNotFound


def remove_dataset(api_client: Client, id: int, team_slug: Optional[str] = None) -> int:
    """
    Creates a new dataset for the given team

    Parameters
    ----------
    api_client : Client
        The client to use to make the request
    id : int
        The name of the dataset to create

    Returns
    -------
    JSONType
    """
    if not team_slug:
        team_slug = api_client.config.default_team

    response = api_client.put(
        f"/datasets/{id}/archive",
        {"team_slug": team_slug},
    )
    assert isinstance(response, dict)

    if "id" not in response:
        raise DatasetNotFound(f"Dataset with id {id} not found")

    return int(response["id"])



from pathlib import Path

import pytest

from darwin.future.data_objects.workflow import WFUser

test_data_path: Path = Path(__file__).parent / "data"
validate_json = test_data_path / "user.json"


def test_file_exists() -> None:
    # This is a test sanity check to make sure the file exists
    # Helps avoids headaches when debugging tests
    assert validate_json.exists()


def test_WFUser_validates_from_valid_json() -> None:
    parsed_user = WFUser.parse_file(validate_json)

    assert isinstance(parsed_user, WFUser)
    assert parsed_user.user_id == 100
    assert str(parsed_user.stage_id) == "0fa1ae43-fb46-44d7-bf85-b78e81d0d02f"



import ast
import json as native_json
from asyncore import loop
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

from rich.console import Console

console = Console()
try:
    import nibabel as nib
    from nibabel.orientations import axcodes2ornt, io_orientation, ornt_transform
except ImportError:
    import_fail_string = """
    You must install darwin-py with pip install darwin-py\[medical]
    in order to export using nifti format
    """
    console.print(import_fail_string)
    exit()
import numpy as np
import orjson as json
from PIL import Image

import darwin.datatypes as dt
from darwin.utils import convert_polygons_to_mask, get_progress_bar


@dataclass
class Volume:
    pixel_array: np.ndarray
    affine: Optional[np.ndarray]
    original_affine: Optional[np.ndarray]
    dims: List
    pixdims: List
    class_name: str
    series_instance_uid: str


def export(annotation_files: Iterable[dt.AnnotationFile], output_dir: Path) -> None:
    """
    Exports the given ``AnnotationFile``\\s into nifti format inside of the given
    ``output_dir``. Deletes everything within ``output_dir/masks`` before writting to it.

    Parameters
    ----------
    annotation_files : Iterable[dt.AnnotationFile]
        The ``AnnotationFile``\\s to be exported.
    output_dir : Path
        The folder where the new instance mask files will be.

    Returns
    -------
    sends output volumes, image_id and output_dir to the write_output_volume_to_disk function
    
    """

    
    video_annotations = list(annotation_files)    
    for video_annotation in video_annotations:
        image_id = check_for_error_and_return_imageid(video_annotation, output_dir)
        if not isinstance(image_id, str):
            continue
        output_volumes = build_output_volumes(video_annotation)
        slot_map = {slot.name:slot for slot in video_annotation.slots}
        for annotation in video_annotation.annotations:
            #print(slot)
            populate_output_volumes(annotation, output_dir, slot_map, output_volumes, image_id)
        write_output_volume_to_disk(output_volumes, image_id=image_id, output_dir=output_dir)

def build_output_volumes(video_annotation: dt.AnnotationFile):
    """
    This is a function to create the output volumes based on the whole annotation file
    
    Parameters
    ----------
    video_annotation : dt.AnnotationFile

    Returns
    -------
    output_volumes: Dict    
        The output volume built per class

    """   
    # Builds a map of class to integer
    class_map = {}
    class_count = 1
    for annotation in video_annotation.annotations:
        frames = annotation.frames
        for frame_idx in frames.keys():
            class_name = frames[frame_idx].annotation_class.name
            if class_name not in class_map:
                class_map[class_name] = class_count
                class_count += 1

    output_volumes = {}
    for slot in video_annotation.slots:
        slot_metadata = slot.metadata
        series_instance_uid = slot_metadata.get("SeriesInstanceUID", "SeriesIntanceUIDNotProvided")
        # Builds output volumes per class
        volume_dims, pixdims, affine, original_affine = process_metadata(slot.metadata)
        output_volumes[series_instance_uid] = {class_name: 
                                               Volume(pixel_array=np.zeros(volume_dims),
                                                      affine=affine,
                                                      original_affine=original_affine,
                                                      dims=volume_dims,
                                                      pixdims=pixdims,
                                                      series_instance_uid=series_instance_uid,
                                                      class_name=class_name) for class_name in class_map.keys()}
    return output_volumes


def check_for_error_and_return_imageid(video_annotation: dt.AnnotationFile, output_dir: Path):
    """
    Given the video_annotation file and the output directory, checks for a range of errors and 
    returns messages accordingly.

    Parameters
    ----------
    video_annotation : dt.AnnotationFile
        The ``AnnotationFile``\\s to be exported.
    output_dir : Path
        The folder where the new instance mask files will be.

    Returns
    -------
    image_id : str
    
    """


    output_volumes = None
    filename = Path(video_annotation.filename)
    suffixes = filename.suffixes
    if len(suffixes) > 2:
        return create_error_message_json(
            "Misconfigured filename, contains too many suffixes", output_dir, str(filename)
        )
    elif len(suffixes) == 2:
        if suffixes[0] == ".nii" and suffixes[1] == ".gz":
            image_id = str(filename).strip("".join(suffixes))
        else:
            return create_error_message_json("Two suffixes found but not ending in .nii.gz", output_dir, str(filename))
    elif len(suffixes) == 1:
        if suffixes[0] == ".nii" or suffixes[0] == ".dcm":
            image_id = filename.stem
        else:
            return create_error_message_json(
                "Misconfigured filename, not ending in .nii or .dcm. Are you sure this is medical data?",
                output_dir,
                str(filename),
            )
    else:
        return create_error_message_json(
            "You are trying to export to nifti. Filename should contain either .nii, .nii.gz or .dcm extension."
            "Are you sure this is medical data?",
            output_dir,
            str(filename),
        )
    if video_annotation is None:
        return create_error_message_json("video_annotation not found", output_dir, image_id)  
    
    for slot in video_annotation.slots:
        # Pick the first slot to take the metadata from. We assume that all slots have the same metadata.
        metadata = slot.metadata
        if metadata is None:
            return create_error_message_json(
                f"No metadata found for {str(filename)}, are you sure this is medical data?", output_dir, image_id
            )

        volume_dims, pixdim, affine, _ = process_metadata(metadata)
        if affine is None or pixdim is None or volume_dims is None:
            return create_error_message_json(
                f"Missing one of affine, pixdim or shape in metadata for {str(filename)}, try reuploading file",
                output_dir,
                image_id,
            )
    return image_id


def populate_output_volumes(annotation: Union[dt.Annotation, dt.VideoAnnotation], output_dir: str, slot_map: Dict, output_volumes: Dict, image_id: str) -> None:

    """
    Exports the given ``AnnotationFile``\\s into nifti format inside of the given
    ``output_dir``. Deletes everything within ``output_dir/masks`` before writting to it.

    Parameters
    ----------
    annotation : Union[dt.Annotation, dt.VideoAnnotation]
        The Union of these two files used to populate the volume with
    output_dir : Path
        The folder where the new instance mask files will be.
    slot_map : Dict
        Dictionary of the different slots within the annotation file
    output_volumes : Dict
        volumes created from the build_output_volumes file
    image_id : str

    Returns
    -------
    volume : dict
        Returns the populated volume
    
    """

    slot_name = annotation.slot_names[0]
    slot = slot_map[slot_name]
    series_instance_uid = slot.metadata.get("SeriesInstanceUID", "SeriesIntanceUIDNotProvided")
    volume = output_volumes.get(series_instance_uid)
    frames = annotation.frames
    frame_new = {}  

    # define the different planes
    XYPLANE = 0
    XZPLANE = 1
    YZPLANE = 2 
          
    for frame_idx in frames.keys():
        frame_new[frame_idx] = frames
        view_idx = get_view_idx_from_slot_name(slot_name, slot.metadata.get("orientation"))
        if view_idx == XYPLANE:
            height, width = volume[annotation.annotation_class.name].dims[0], volume[annotation.annotation_class.name].dims[1]
        elif view_idx == XZPLANE:
            height, width = volume[annotation.annotation_class.name].dims[0], volume[annotation.annotation_class.name].dims[2]
        elif view_idx == YZPLANE:
            height, width = volume[annotation.annotation_class.name].dims[1], volume[annotation.annotation_class.name].dims[2]
        if "paths" in frames[frame_idx].data:
            # Dealing with a complex polygon
            polygons = [shift_polygon_coords(polygon_path, volume[annotation.annotation_class.name].pixdims) for polygon_path in frames[frame_idx].data["paths"]]
        elif "path" in frames[frame_idx].data:
            # Dealing with a simple polygon
            polygons = shift_polygon_coords(frames[frame_idx].data["path"], volume[annotation.annotation_class.name].pixdims)
        else:
            continue
        class_name = frames[frame_idx].annotation_class.name
        im_mask = convert_polygons_to_mask(polygons, height=height, width=width)
        volume = output_volumes[series_instance_uid]
        if view_idx == 0:
            volume[annotation.annotation_class.name].pixel_array[:, :, frame_idx] = np.logical_or(im_mask, volume[annotation.annotation_class.name].pixel_array[:, :, frame_idx])
        elif view_idx == 1:
            volume[annotation.annotation_class.name].pixel_array[:, frame_idx, :] = np.logical_or(im_mask, volume[annotation.annotation_class.name].pixel_array[:, frame_idx, :])
        elif view_idx == 2:
            volume[annotation.annotation_class.name].pixel_array[frame_idx, :, :] = np.logical_or(im_mask, volume[annotation.annotation_class.name].pixel_array[frame_idx, :, :])


def write_output_volume_to_disk(output_volumes: Dict, image_id: str, output_dir: str):
    # volumes are the values of this nested dict
    def unnest_dict_to_list(d):
        result = []
        for value in d.values():
            if isinstance(value, dict):
                result.extend(unnest_dict_to_list(value))
            else:
                result.append(value)
        return result
    volumes = unnest_dict_to_list(output_volumes)
    for volume in volumes:
        img = nib.Nifti1Image(
            dataobj=np.flip(volume.pixel_array, (0, 1, 2)).astype(np.int16),
            affine=volume.affine,
        )
        if volume.original_affine is not None:
            orig_ornt = io_orientation(volume.original_affine)  # Get orientation of current affine
            img_ornt = io_orientation(volume.affine) # Get orientation of RAS affine
            from_canonical = ornt_transform(img_ornt, orig_ornt)  # Get transform from RAS to current affine
            img = img.as_reoriented(from_canonical)
        output_path = Path(output_dir) / f"{image_id}_{volume.series_instance_uid}_{volume.class_name}.nii.gz"
        if not output_path.parent.exists():
            output_path.parent.mkdir(parents=True)
        nib.save(img=img, filename=output_path)

def shift_polygon_coords(polygon, pixdim):
    # Need to make it clear that we flip x/y because we need to take the transpose later.
    if pixdim[1] > pixdim[0]:
        return [{"x": p["y"], "y": p["x"] * pixdim[1] / pixdim[0]} for p in polygon]
    elif pixdim[1] < pixdim[0]:
        return [{"x": p["y"] * pixdim[0] / pixdim[1], "y": p["x"]} for p in polygon]
    else:
        return [{"x": p["y"], "y": p["x"]} for p in polygon]
    

def get_view_idx(frame_idx, groups):
    if groups is None:
        return 0
    for view_idx, group in enumerate(groups):
        if frame_idx in group:
            return view_idx


def get_view_idx_from_slot_name(slot_name, orientation):
    if orientation is None:
        orientation_dict = {"0.1": 0, "0.2": 1, "0.3": 2}
        return orientation_dict.get(slot_name, 0)
    else:
        orientation_dict = {"AXIAL": 0, "SAGITTAL": 1, "CORONAL": 2}
        return orientation_dict.get(orientation, 0)


def process_metadata(metadata):
    volume_dims = metadata.get("shape")
    pixdim = metadata.get("pixdim")
    affine = process_affine(metadata.get("affine"))
    original_affine = process_affine(metadata.get("original_affine"))
    # If the original affine is in the medical payload of metadata then use it
    if isinstance(pixdim, str):
        pixdim = ast.literal_eval(pixdim)
        if isinstance(pixdim, tuple) or isinstance(pixdim, list):
            if len(pixdim) == 4:
                pixdim = pixdim[1:]
            if len(pixdim) != 3:
                pixdim = None
        else:
            pixdim = None
    if isinstance(volume_dims, list):
        if volume_dims:
            if volume_dims[0] == 1:  # remove first singleton dimension
                volume_dims = volume_dims[1:]
        else:
            volume_dims = None
    return volume_dims, pixdim, affine, original_affine


def process_affine(affine):
    if isinstance(affine, str):
        affine = np.squeeze(np.array([ast.literal_eval(l) for l in affine.split("\n")]))
    elif isinstance(affine, list):
        affine = np.array(affine).astype(float)
    else:
        return
    if isinstance(affine, np.ndarray):
        return affine


def create_error_message_json(error_message: str, output_dir: str, image_id: str, slot_name: str):
    output_path = Path(output_dir) / f"{image_id}_{slot_name}_error.json"
    if not output_path.parent.exists():
        output_path.parent.mkdir(parents=True)
    with open(output_path, "w") as f:
        native_json.dump({"error": error_message}, f)

    return False





from typing import List

import pytest
import responses

from darwin.future.core.client import Client
from darwin.future.data_objects.workflow import WFType, Workflow
from darwin.future.meta.objects.stage import StageMeta
from darwin.future.meta.objects.workflow import WorkflowMeta
from darwin.future.meta.queries.stage import StageQuery
from darwin.future.tests.core.fixtures import *


@pytest.fixture
def filled_query(base_client: Client, base_workflow_meta: WorkflowMeta) -> StageQuery:
    return StageQuery(base_client, meta_params={"workflow_id": str(base_workflow_meta.id)})


@pytest.fixture
def base_workflow_meta(base_client: Client, base_single_workflow_object: dict) -> WorkflowMeta:
    return WorkflowMeta(base_client, Workflow.parse_obj(base_single_workflow_object))


@pytest.fixture
def multi_stage_workflow_object(base_single_workflow_object: dict) -> dict:
    stage = base_single_workflow_object["stages"][0]
    types = [t for t in WFType.__members__.values()] * 3
    stages = []
    for i, t in enumerate(types):
        temp = stage.copy()
        temp["name"] = f"stage{i}"
        temp["type"] = t.value
        stages.append(temp)
    base_single_workflow_object["stages"] = stages
    return base_single_workflow_object


def test_stage_collects_basic(
    filled_query: StageQuery, base_single_workflow_object: dict, base_workflow_meta: WorkflowMeta
) -> None:
    UUID = base_workflow_meta.id
    with responses.RequestsMock() as rsps:
        endpoint = filled_query.client.config.api_endpoint + f"v2/teams/default-team/workflows/{UUID}"
        rsps.add(responses.GET, endpoint, json=base_single_workflow_object)
        stages = filled_query.collect()
        assert len(stages) == len(base_workflow_meta.stages)
        assert isinstance(stages[0], StageMeta)


def test_stage_filters_basic(
    filled_query: StageQuery, multi_stage_workflow_object: dict, base_workflow_meta: WorkflowMeta
) -> None:
    UUID = base_workflow_meta.id
    with responses.RequestsMock() as rsps:
        endpoint = filled_query.client.config.api_endpoint + f"v2/teams/default-team/workflows/{UUID}"
        rsps.add(responses.GET, endpoint, json=multi_stage_workflow_object)
        stages = filled_query.where({"name": "name", "param": "stage1"}).collect()
        assert len(stages) == 1
        assert isinstance(stages[0], StageMeta)
        assert stages[0]._item is not None
        assert stages[0]._item.name == "stage1"


@pytest.mark.parametrize("wf_type", [t for t in WFType.__members__.values()])
def test_stage_filters_WFType(
    wf_type: WFType, filled_query: StageQuery, multi_stage_workflow_object: dict, base_workflow_meta: WorkflowMeta
) -> None:
    UUID = base_workflow_meta.id
    with responses.RequestsMock() as rsps:
        endpoint = filled_query.client.config.api_endpoint + f"v2/teams/default-team/workflows/{UUID}"
        rsps.add(responses.GET, endpoint, json=multi_stage_workflow_object)
        stages = filled_query.where({"name": "type", "param": wf_type.value}).collect()
        assert len(stages) == 3
        assert isinstance(stages[0], StageMeta)
        for stage in stages:
            assert stage._item is not None
            assert stage._item.type == wf_type



from darwin.future.core.types.common import QueryString
from pytest import raises


# happy and sad path tests for QueryString - should validate a dict of strings, and return a query string on str()
def test_querystring_happy_path() -> None:
    query_string = QueryString({"foo": "bar"})
    assert str(query_string) == "?foo=bar"

    query_string_2 = QueryString({"foo": "bar", "baz": "qux"})
    assert str(query_string_2) == "?foo=bar&baz=qux"

    query_string_3 = QueryString(dict())
    assert str(query_string_3) == "?"

    assert query_string.value == {"foo": "bar"}
    assert query_string_2.value == {"foo": "bar", "baz": "qux"}


def test_querystring_sad_path() -> None:
    with raises(AssertionError):
        QueryString({"foo": 1})  # type: ignore



from typing import List

import responses
from requests.exceptions import HTTPError

from darwin.future.core.client import Client
from darwin.future.core.datasets.list_datasets import list_datasets
from darwin.future.core.types import TeamSlug
from darwin.future.core.types.common import QueryString
from darwin.future.data_objects.dataset import Dataset
from darwin.future.tests.core.fixtures import *

from .fixtures import *


def test_it_lists_datasets(base_client: Client, basic_list_of_datasets: List[Dataset]) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint + "datasets",
            json=basic_list_of_datasets,
            status=200,
        )

        datasets, errors = list_datasets(base_client)

        assert len(errors) == 0

        assert len(datasets) == 3
        assert datasets[0].name == "test-dataset"
        assert datasets[0].slug == "1337"


def test_it_returns_an_error_if_the_client_returns_an_http_error(base_client: Client) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint + "datasets",
            json={},
            status=400,
        )

        response, errors = list_datasets(base_client)

        assert len(errors) == 1
        assert isinstance(error := errors[0], HTTPError)
        assert error.response.status_code == 400
        assert not response



import unittest
from typing import Any, List, Optional, Type

import pytest

from darwin.future.core.client import Client
from darwin.future.core.types import query as Query
from darwin.future.data_objects.team import Team
from darwin.future.tests.core.fixtures import *


@pytest.fixture
def non_abc_query() -> Type[Query.Query]:
    """Query is an abstract base class, this fixture removes the abstract methods
    so that it can be instantiated for testing the default methods
    """
    fixed = Query.Query
    fixed.__abstractmethods__ = set()  # type: ignore
    return fixed


@pytest.fixture
def basic_filters() -> List[Query.QueryFilter]:
    return [
        Query.QueryFilter(name="test1", param="test1"),
        Query.QueryFilter(name="test2", param="test2"),
        Query.QueryFilter(name="test3", param="test3"),
    ]


@pytest.fixture
def test_team() -> Team:
    return Team(slug="test-team", id=0)


def test_query_instantiated(
    base_client: Client, basic_filters: List[Query.QueryFilter], non_abc_query: Type[Query.Query]
) -> None:
    q = non_abc_query(base_client, basic_filters)
    assert q.filters == basic_filters


def test_query_filter_functionality(
    base_client: Client, basic_filters: List[Query.QueryFilter], non_abc_query: Type[Query.Query]
) -> None:
    q = non_abc_query(base_client)
    for f in basic_filters:
        q = q.filter(f)
    assert q.filters == basic_filters
    dropped_filter = basic_filters.pop(0)
    q = q - dropped_filter
    assert q.filters == basic_filters
    basic_filters.append(dropped_filter)
    q = q + dropped_filter
    assert q.filters == basic_filters

    # Test filter drops from middle
    dropped_filter = basic_filters.pop(1)
    q = q - dropped_filter
    assert q.filters == basic_filters
    basic_filters.append(dropped_filter)
    q = q + dropped_filter
    assert q.filters == basic_filters


@pytest.mark.parametrize(
    "mod,param,check,expected",
    [
        (None, "test", "test", True),  # test str equalities
        ("!=", "test", "test", False),
        ("contains", "test", "test", True),
        (None, "test", "test1", False),  # test str inequalites
        ("!=", "test", "test1", True),
        ("contains", "test1", "test", False),
        (None, 1, 1, True),  # test int equalities
        ("!=", 1, 1, False),
        (">", 1, 2, True),
        (">=", 1, 1, True),
        ("<", 1, 1, False),
        ("<=", 1, 1, True),
        (">", 1, 1, False),  # test int inequalites
        ("<", 1, 2, False),
        (">=", 1, 0, False),
        ("<=", 1, 2, False),
        (None, 1, 2, False),
        ("!=", 1, 2, True),
    ],
)
def test_query_filter_filters(mod: Optional[str], param: Any, check: Any, expected: bool) -> None:  # type: ignore
    # test str
    if mod:
        modifier = Query.Modifier(mod)
    else:
        modifier = None
    QF = Query.QueryFilter(name="test", param=param, modifier=modifier)
    assert QF.filter_attr(check) == expected


def test_QF_from_asteriks() -> None:
    # Builds with dictionary args
    QF = Query.QueryFilter._from_args(
        {"name": "test", "param": "test"}, {"name": "test2", "param": "test2", "modifier": "!="}
    )
    assert len(QF) == 2
    assert QF[0].name == "test"
    assert QF[0].param == "test"
    assert QF[0].modifier is None
    assert QF[1].name == "test2"
    assert QF[1].param == "test2"
    assert QF[1].modifier == Query.Modifier("!=")

    # Builds with kwargs
    QF = Query.QueryFilter._from_args(test="test", test2="!=:test2")
    assert len(QF) == 2
    assert QF[0].name == "test"
    assert QF[0].param == "test"
    assert QF[0].modifier is None
    assert QF[1].name == "test2"
    assert QF[1].param == "test2"
    assert QF[1].modifier == Query.Modifier("!=")

    # fails on bad args
    with pytest.raises(ValueError):
        Query.QueryFilter._from_args({})
        Query.QueryFilter._from_args([])
        Query.QueryFilter._from_args(1, 2, 3)



from pytest import fixture, raises

from darwin.future.core.client import DarwinConfig
from darwin.future.meta.client import MetaClient
from darwin.future.tests.core.fixtures import *


@fixture
def base_meta_client(base_config: DarwinConfig) -> MetaClient:
    return MetaClient(base_config)


from pathlib import Path
from typing import List

import orjson as json
import pytest

from darwin.future.core.client import Client, DarwinConfig
from darwin.future.data_objects.dataset import Dataset
from darwin.future.data_objects.team import Team, TeamMember
from darwin.future.data_objects.team_member_role import TeamMemberRole
from darwin.future.data_objects.workflow import WFType


@pytest.fixture
def base_config() -> DarwinConfig:
    return DarwinConfig(
        api_key="test_key",
        base_url="http://test_url.com/",
        api_endpoint="http://test_url.com/api/",
        default_team="default-team",
        datasets_dir=Path("datasets"),
        teams={},
    )


@pytest.fixture
def base_client(base_config: DarwinConfig) -> Client:
    return Client(base_config)


@pytest.fixture
def base_team_json() -> dict:
    return {"slug": "test-team", "id": "0"}


@pytest.fixture
def base_team(base_team_json: dict) -> Team:
    return Team.parse_obj(base_team_json)


@pytest.fixture
def base_team_member_json() -> dict:
    return {
        "email": "email",
        "id": "0",
        "first_name": "first",
        "last_name": "last",
        "role": "member",
        "team_id": "0",
        "user_id": "0",
    }


@pytest.fixture
def base_team_member(base_team_member_json: dict) -> TeamMember:
    return TeamMember.parse_obj(base_team_member_json)


@pytest.fixture
def base_team_members_json(base_team_member_json: dict) -> List[dict]:
    members = []
    for item in TeamMemberRole:
        member_w_role = base_team_member_json.copy()
        member_w_role["role"] = item.value
        members.append(member_w_role)
    return members


@pytest.fixture
def team_members(base_team_members_json: List[dict]) -> List[TeamMember]:
    return [TeamMember.parse_obj(item) for item in base_team_members_json]


@pytest.fixture
def base_dataset_json() -> dict:
    return {
        "name": "Test Dataset",
        "slug": "test-dataset",
        "id": "1",
        "releases": [],
    }


@pytest.fixture
def base_dataset_json_with_releases() -> dict:
    dataset = base_dataset_json()
    dataset["releases"] = [
        {"name": "release1"},
        {"name": "release2"},
    ]

    return dataset


@pytest.fixture
def base_dataset(base_dataset_json: dict) -> Dataset:
    return Dataset.parse_obj(base_dataset_json)


def base_dataset_with_releases(base_dataset_json_with_releases: dict) -> Dataset:
    return Dataset.parse_obj(base_dataset_json_with_releases)


@pytest.fixture
def base_datasets_json(base_dataset_json: dict) -> List[dict]:
    def transform_dataset(dataset_json_dict: dict, id: int) -> dict:
        dataset = dataset_json_dict.copy()

        dataset["id"] = id
        dataset["slug"] = f"{dataset['slug']}-{id}"
        dataset["name"] = f"{dataset['name']} {id}"

        return dataset

    # fmt: off
    return [
        transform_dataset(base_dataset_json, 1),
        transform_dataset(base_dataset_json, 2),
    ]
    # fmt: on


@pytest.fixture
def base_datasets_json_with_releases(base_dataset_json: dict) -> List[dict]:
    def transform_dataset(dataset_json_dict: dict, id: int) -> dict:
        dataset = dataset_json_dict.copy()

        dataset["id"] = id
        dataset["slug"] = f"{dataset['slug']}-{id}"
        dataset["name"] = f"{dataset['name']} {id}"
        dataset["releases"] = [{"name": "release2"}] if id % 2 == 0 else [{"name": "release1"}]

        return dataset

    # fmt: off
    return [
        transform_dataset(base_dataset_json, 1),
        transform_dataset(base_dataset_json, 2),
        transform_dataset(base_dataset_json, 3),
        transform_dataset(base_dataset_json, 4),
    ]
    # fmt: on


@pytest.fixture
def workflow_json() -> str:
    # fmt: off
    path = (
        Path(__file__).parent / 
        ".." / "data_objects" / 
        "workflow" / "data" / "workflow.json"
    ).resolve()
    # fmt: on
    assert path.exists()

    return path.read_bytes().decode("utf-8").replace("\n", "")


@pytest.fixture
def base_single_workflow_json(workflow_json: str) -> str:
    return f"[{workflow_json}]"


@pytest.fixture
def base_workflows_json(workflow_json: str) -> str:
    return f"[{workflow_json}, {workflow_json}, {workflow_json}]"


@pytest.fixture
def base_workflows_object(base_workflows_json: str) -> list:
    return json.loads(base_workflows_json)


@pytest.fixture
def base_single_workflow_object(base_workflows_object: dict) -> list:
    return base_workflows_object[0]


@pytest.fixture
def base_filterable_workflows() -> list:
    return [
        {
            "id": "6dca86a3-48fb-40cc-8594-88310f5f1fdf",
            "name": "test-workflow-1",
            "team_id": "100",
            "inserted_at": "2021-06-01T15:00:00.000+00:00",
            "updated_at": "2021-06-05T15:00:00.000+00:00",
            "dataset": {
                "id": 1,
                "name": "test-dataset-1",
                "instructions": "test-instructions-1",
            },
            "stages": [
                {
                    "id": "53d2c997-6bb0-4766-803c-3c8d1fb21072",
                    "name": "stage-1",
                    "type": "annotate",
                    "assignable_users": [
                        {
                            "stage_id": "d96a4865-f7d1-466a-abc6-5a61b2339c16",
                            "user_id": "1",
                        },
                        {
                            "stage_id": "70adf8be-d6b5-4f54-9d99-ecf1c6959442",
                            "user_id": "2",
                        },
                    ],
                    "edges": [
                        {
                            "id": "5e858c07-28d7-48b5-a7a3-4697f3212d7c",
                            "name": "edge-1",
                            "source_stage_id": None,
                            "target_stage_id": "6aeb1b33-9234-4d00-95e7-97b8e477ee02",
                        },
                        {
                            "id": "6aeb1b33-9234-4d00-95e7-97b8e477ee02",
                            "name": "edge-2",
                            "source_stage_id": "5e858c07-28d7-48b5-a7a3-4697f3212d7c",
                            "target_stage_id": "32151eaf-edbd-4703-9049-50803f1df2bf",
                        },
                        {
                            "id": "9b527a3d-c765-42fd-88b6-594f5b411c07",
                            "name": "edge-3",
                            "source_stage_id": "6aeb1b33-9234-4d00-95e7-97b8e477ee02",
                            "target_stage_id": None,
                        },
                    ],
                }
            ],
            "thumbnails": [
                "https://0.0.0.0/thumbnails/1.png",
                "https://0.0.0.0/thumbnails/2.png",
                "https://0.0.0.0/thumbnails/3.png",
            ],
        },
        {
            "id": "e34fe935-4a1c-4231-bb55-454e2ac7673f",
            "name": "test-workflow-2",
            "team_id": "100",
            "inserted_at": "2021-06-03T15:00:00.000+00:00",
            "updated_at": "2021-06-05T15:00:00.000+00:00",
            "dataset": {
                "id": 2,
                "name": "test-dataset-2",
                "instructions": "test-instructions-1",
            },
            "stages": [
                {
                    "id": "aabd5e3a-4ccd-4cc3-8cc1-c8455972c101",
                    "name": "stage-1",
                    "type": "annotate",
                    "assignable_users": [
                        {
                            "stage_id": "d96a4865-f7d1-466a-abc6-5a61b2339c16",
                            "user_id": "1",
                        },
                        {
                            "stage_id": "70adf8be-d6b5-4f54-9d99-ecf1c6959442",
                            "user_id": "2",
                        },
                    ],
                    "edges": [
                        {
                            "id": "42c53da7-e3f2-4c81-bec7-449439cef694",
                            "name": "edge-1",
                            "source_stage_id": None,
                            "target_stage_id": "b5da9e56-4bf4-4a00-826d-741d1febd3da",
                        },
                        {
                            "id": "8ea42761-8971-4be1-b359-66fc878a807b",
                            "name": "edge-2",
                            "source_stage_id": "b5da9e56-4bf4-4a00-826d-741d1febd3da",
                            "target_stage_id": "2a3d8f47-dba6-4fdb-88e0-beb4b2a5ed24",
                        },
                        {
                            "id": "2a3d8f47-dba6-4fdb-88e0-beb4b2a5ed24",
                            "name": "edge-3",
                            "source_stage_id": "8ea42761-8971-4be1-b359-66fc878a807b",
                            "target_stage_id": None,
                        },
                    ],
                }
            ],
            "thumbnails": [
                "https://0.0.0.0/thumbnails/4.png",
                "https://0.0.0.0/thumbnails/5.png",
                "https://0.0.0.0/thumbnails/6.png",
            ],
        },
        {
            "id": "45cf0abe-58a2-4878-b171-4fb5421a1c39",
            "name": "test-workflow-3",
            "team_id": "100",
            "inserted_at": "2021-06-05T15:00:00.000+00:00",
            "updated_at": "2021-06-10T15:00:00.000+00:00",
            "dataset": {
                "id": 3,
                "name": "test-dataset-3",
                "instructions": "test-instructions-1",
            },
            "stages": [
                {
                    "id": "5445adcb-193d-4f76-adb0-0c6d5f5e4c04",
                    "name": "stage-1",
                    "type": "annotate",
                    "assignable_users": [
                        {
                            "stage_id": "16d390ad-2a6b-4232-8434-0489e9533afd",
                            "user_id": "1",
                        },
                        {
                            "stage_id": "f9816ea0-4bcf-4293-a399-3d35770db59e",
                            "user_id": "2",
                        },
                    ],
                    "edges": [
                        {
                            "id": "7dc64300-dc1b-42f3-825f-de09ece4ed6f",
                            "name": "edge-1",
                            "source_stage_id": None,
                            "target_stage_id": "d7510016-5286-4f72-a13c-e350fefb652b",
                        },
                        {
                            "id": "d7510016-5286-4f72-a13c-e350fefb652b",
                            "name": "edge-2",
                            "source_stage_id": "7dc64300-dc1b-42f3-825f-de09ece4ed6f",
                            "target_stage_id": "7233d3e4-ac01-4598-aada-1fc73e6fc518",
                        },
                        {
                            "id": "7233d3e4-ac01-4598-aada-1fc73e6fc518",
                            "name": "edge-3",
                            "source_stage_id": "d7510016-5286-4f72-a13c-e350fefb652b",
                            "target_stage_id": None,
                        },
                    ],
                }
            ],
            "thumbnails": [
                "https://0.0.0.0/thumbnails/7.png",
                "https://0.0.0.0/thumbnails/8.png",
                "https://0.0.0.0/thumbnails/9.png",
            ],
        },
    ]



from typing import List
from uuid import UUID

from pytest import fixture, raises

from darwin.future.core.client import Client
from darwin.future.data_objects.team import Team
from darwin.future.data_objects.workflow import WFStage, Workflow
from darwin.future.meta.objects import stage
from darwin.future.meta.objects.stage import StageMeta
from darwin.future.meta.objects.team import TeamMeta
from darwin.future.meta.objects.workflow import WorkflowMeta
from darwin.future.tests.core.fixtures import *


@fixture
def base_UUID() -> UUID:
    return UUID("00000000-0000-0000-0000-000000000000")


@fixture
def base_meta_team(base_client: Client, base_team: Team) -> TeamMeta:
    return TeamMeta(base_client, base_team)


@fixture
def base_meta_workflow(base_client: Client, base_workflow: Workflow) -> WorkflowMeta:
    return WorkflowMeta(base_client, base_workflow)


@fixture
def base_meta_stage(base_client: Client, base_stage: WFStage, base_UUID: UUID) -> StageMeta:
    return StageMeta(base_client, base_stage, base_UUID)


@fixture
def base_meta_stage_list(base_meta_stage: StageMeta, base_UUID: UUID) -> List[StageMeta]:
    return [base_meta_stage]



from pydantic import parse_obj_as

from darwin.future.core.client import Client
from darwin.future.data_objects.dataset import Dataset


def create_dataset(api_client: Client, name: str) -> Dataset:
    """
    Creates a new dataset for the given team

    Parameters
    ----------

    api_client: Client
        The client to use to make the request
    name: str
        The name of the dataset to create

    Returns
    -------
    Dataset
        The created dataset

    Raises
    ------
    HTTPError
        Any HTTP errors returned by the API
    """
    response = api_client.post(
        "/datasets",
        {
            "name": name,
        },
    )

    return parse_obj_as(Dataset, response)



from typing import List, Optional

from pydantic import parse_obj_as

from darwin.future.core.client import Client
from darwin.future.data_objects.workflow import Workflow


def get_workflows(client: Client, team_slug: Optional[str] = None) -> List[Workflow]:
    team_slug = team_slug or client.config.default_team
    response = client.get(f"/v2/teams/{team_slug}/workflows?worker=false")

    return [parse_obj_as(Workflow, workflow) for workflow in response]



from __future__ import annotations

from typing import List
from uuid import UUID

from darwin.future.core.client import Client
from darwin.future.core.types.query import Param, Query, QueryFilter
from darwin.future.core.workflows.get_workflow import get_workflow
from darwin.future.meta.objects.stage import StageMeta


class StageQuery(Query[StageMeta]):
    def collect(self) -> List[StageMeta]:
        if not self.meta_params:
            raise ValueError("Must specify workflow_id to query stages")
        workflow_id: UUID = self.meta_params["workflow_id"]
        workflow, exceptions = get_workflow(self.client, str(workflow_id))
        assert workflow is not None
        stages = [StageMeta(self.client, s, workflow_id) for s in workflow.stages]
        if not self.filters:
            self.filters = []
        for filter in self.filters:
            stages = self._execute_filter(stages, filter)
        return stages

    def _execute_filter(self, stages: List[StageMeta], filter: QueryFilter) -> List[StageMeta]:
        """Executes filtering on the local list of stages
        Parameters
        ----------
        stages : List[Stage]
        filter : QueryFilter

        Returns
        -------
        List[Stage]: Filtered subset of stages
        """
        if filter.name == "role":
            return [s for s in stages if s._item is not None and filter.filter_attr(s._item.type.value)]
        return super()._generic_execute_filter(stages, filter)



from typing import Dict

from darwin.future.data_objects import validators as darwin_validators
from pydantic import BaseModel
import pydantic

from darwin.future.data_objects.typing import UnknownType


class TeamSlug(str):
    """Team slug type"""

    min_length = 1
    max_length = 256

    @classmethod
    def __get_validators__(cls):  # type: ignore
        yield cls.validate

    @classmethod
    def validate(cls, v: str) -> "TeamSlug":
        assert len(v) < cls.max_length, f"maximum length for team slug is {cls.max_length}"
        assert len(v) > cls.min_length, f"minimum length for team slug is {cls.min_length}"
        if not isinstance(v, str):
            raise TypeError("string required")
        modified_value = darwin_validators.parse_name(v)
        return cls(modified_value)

    def __repr__(self) -> str:
        return f"TeamSlug({super().__repr__()})"


class QueryString:
    """Query string type"""

    value: Dict[str, str]

    def dict_check(cls, value: UnknownType) -> Dict[str, str]:
        assert isinstance(value, dict)
        assert all(isinstance(k, str) and isinstance(v, str) for k, v in value.items())
        return value

    def __init__(self, value: Dict[str, str]) -> None:
        self.value = self.dict_check(value)

    def __str__(self) -> str:
        return "?" + "&".join(f"{k}={v}" for k, v in self.value.items())



from __future__ import annotations

from typing import List

from darwin.future.core.datasets.list_datasets import list_datasets
from darwin.future.core.types.query import Param, Query, QueryFilter
from darwin.future.data_objects.dataset import Dataset
from darwin.future.meta.objects.dataset import DatasetMeta


class DatasetQuery(Query[DatasetMeta]):
    """
    DatasetQuery object with methods to manage filters, retrieve data, and execute
    filters

    Methods
    -------

    collect: Executes the query and returns the filtered data
    """

    def collect(self) -> List[DatasetMeta]:
        datasets, exceptions = list_datasets(self.client)
        if exceptions:
            # TODO: print and or raise exceptions, tbd how we want to handle this
            pass
        datasets_meta = [DatasetMeta(self.client, dataset) for dataset in datasets]
        if not self.filters:
            self.filters = []

        for filter in self.filters:
            datasets_meta = self._execute_filters(datasets_meta, filter)

        return datasets_meta

    def _execute_filters(self, datasets: List[DatasetMeta], filter: QueryFilter) -> List[DatasetMeta]:
        """Executes filtering on the local list of datasets, applying special logic for role filtering
        otherwise calls the parent method for general filtering on the values of the datasets

        Parameters
        ----------
        datasets : List[Dataset]
        filter : QueryFilter

        Returns
        -------
        List[Dataset]: Filtered subset of datasets
        """

        if filter.name == "releases":
            return [
                d
                for d in datasets
                if d._item is not None and d._item.releases and filter.param in [str(r) for r in d._item.releases]
            ]

        return super()._generic_execute_filter(datasets, filter)



import responses
from pytest import fixture, mark

from darwin.future.core.client import Client
from darwin.future.data_objects.dataset import Dataset
from darwin.future.meta.objects.dataset import DatasetMeta
from darwin.future.meta.queries.dataset import DatasetQuery
from darwin.future.tests.core.fixtures import *


def test_dataset_collects_basic(base_client: Client, base_datasets_json: dict) -> None:
    query = DatasetQuery(base_client)
    with responses.RequestsMock() as rsps:
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json)
        datasets = query.collect()
        assert len(datasets) == 2
        assert all([isinstance(dataset, DatasetMeta) for dataset in datasets])


def test_datasetquery_only_passes_back_correctly_formed_objects(base_client: Client, base_dataset_json: dict) -> None:
    query = DatasetQuery(base_client)
    with responses.RequestsMock() as rsps:
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=[base_dataset_json, {}])
        datasets = query.collect()

        assert len(datasets) == 1
        assert isinstance(datasets[0], DatasetMeta)


def test_dataset_filters_name(base_client: Client, base_datasets_json: dict) -> None:
    with responses.RequestsMock() as rsps:
        query = DatasetQuery(base_client).where({"name": "name", "param": "test dataset 1"})
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json)
        datasets = query.collect()

        assert len(datasets) == 1
        assert datasets[0]._item is not None
        assert datasets[0]._item.slug == "test-dataset-1"


def test_dataset_filters_id(base_client: Client, base_datasets_json: dict) -> None:
    with responses.RequestsMock() as rsps:
        query = DatasetQuery(base_client).where({"name": "id", "param": 1})
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json)
        datasets = query.collect()

        assert len(datasets) == 1
        assert datasets[0]._item is not None
        assert datasets[0]._item.slug == "test-dataset-1"


def test_dataset_filters_slug(base_client: Client, base_datasets_json: dict) -> None:
    with responses.RequestsMock() as rsps:
        query = DatasetQuery(base_client).where({"name": "slug", "param": "test-dataset-1"})
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json)
        datasets = query.collect()

        assert len(datasets) == 1
        assert datasets[0]._item is not None
        assert datasets[0]._item.slug == "test-dataset-1"


def test_dataset_filters_releases(base_client: Client, base_datasets_json_with_releases: dict) -> None:
    with responses.RequestsMock() as rsps:
        query = DatasetQuery(base_client).where({"name": "releases", "param": "release1"})
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json_with_releases)

        datasets_odd_ids = query.collect()

        assert len(datasets_odd_ids) == 2
        assert datasets_odd_ids[0]._item is not None
        assert datasets_odd_ids[1]._item is not None
        assert datasets_odd_ids[0]._item.slug == "test-dataset-1"
        assert datasets_odd_ids[1]._item.slug == "test-dataset-3"

        query2 = DatasetQuery(base_client).where({"name": "releases", "param": "release2"})
        datasets_even_ids = query2.collect()

        assert len(datasets_even_ids) == 2
        assert datasets_even_ids[0]._item is not None
        assert datasets_even_ids[1]._item is not None
        assert datasets_even_ids[0]._item.slug == "test-dataset-2"
        assert datasets_even_ids[1]._item.slug == "test-dataset-4"



from pathlib import Path

import pytest
from pydantic import ValidationError

from darwin.future.data_objects.workflow import WFDataset
from darwin.future.tests.data_objects.workflow.invalidvaluefortest import (
    InvalidValueForTest,
)

test_data_path: Path = Path(__file__).parent / "data"
validate_dataset_json = test_data_path / "dataset.json"


def test_file_exists() -> None:
    # This is a test sanity check to make sure the file exists
    # Helps avoids headaches when debugging tests
    assert validate_dataset_json.exists()


def test_WFDataset_validates_from_valid_json() -> None:
    WFDataset.parse_file(validate_dataset_json)
    assert True


def test_cast_to_int_returns_dataset_id() -> None:
    dataset = WFDataset.parse_file(validate_dataset_json)
    assert dataset.id == 101


def test_cast_to_str_returns_dataset_name() -> None:
    dataset = WFDataset.parse_file(validate_dataset_json)
    assert dataset.name == "Test Dataset"


def test_sad_paths() -> None:
    dataset = WFDataset.parse_file(validate_dataset_json)
    fields = ["id", "name", "instructions"]

    # Test missing fields
    for key in fields:
        with pytest.raises(ValidationError) as excinfo:
            working_dataset = dataset.copy().dict()
            del working_dataset[key]
            WFDataset.parse_obj(working_dataset)

        assert "value_error.missing" in (err_string := str(excinfo.value))
        assert err_string.startswith(f"1 validation error for WFDataset\n{key}")

    # Test invalid types
    for key in fields:
        with pytest.raises(ValidationError) as excinfo:
            working_dataset = dataset.copy().dict()
            working_dataset[key] = InvalidValueForTest()
            WFDataset.parse_obj(working_dataset)

        assert str(excinfo.value).startswith(f"1 validation error for WFDataset\n{key}")



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union, overload
from urllib.parse import urlparse

import requests
import yaml
from pydantic import BaseModel, root_validator, validator
from requests.adapters import HTTPAdapter, Retry

from darwin.future.core.types.common import QueryString
from darwin.future.exceptions.client import NotFound, Unauthorized

JSONType = Union[Dict[str, Any], List[Dict[str, Any]]]  # type: ignore


class TeamsConfig(BaseModel):
    api_key: str
    datasets_dir: Path


class DarwinConfig(BaseModel):
    """Configuration object for the client

    Attributes
    ----------
    api_key: Optional[str], api key to authenticate
    base_url: pydantic.HttpUrl, base url of the API
    default_team: Optional[Team], default team to make requests to
    """

    api_key: Optional[str]
    datasets_dir: Optional[Path]
    api_endpoint: str
    base_url: str
    default_team: str
    teams: Dict[str, TeamsConfig]

    @validator("base_url")
    def validate_base_url(cls, v: str) -> str:
        v = v.strip()
        if not v.endswith("/"):
            v += "/"
        check = urlparse(v)
        assert check.scheme in {"http", "https"}, "base_url must start with http or https"
        assert check.netloc, "base_url must contain a domain"
        return v

    @root_validator(pre=True)
    def remove_global(cls, values: dict) -> dict:
        if "global" not in values:
            return values
        global_conf = values["global"]
        del values["global"]
        values.update(global_conf)
        return values

    @root_validator()
    def validate_defaults(cls, values: dict) -> dict:
        if values["api_key"]:
            return values
        assert values["default_team"] in values["teams"]
        team = values["default_team"]
        values["api_key"] = values["teams"][team].api_key
        values["datasets_dir"] = values["teams"][team].datasets_dir
        return values

    @staticmethod
    def _default_config_path() -> Path:
        return DarwinConfig._default_home_path() / "config.yaml"

    @staticmethod
    def _default_datasets_path() -> Path:
        return DarwinConfig._default_home_path() / "datasets"

    @staticmethod
    def _default_home_path() -> Path:
        return Path.home() / ".darwin"

    @staticmethod
    def local() -> DarwinConfig:
        return DarwinConfig.from_file(DarwinConfig._default_config_path())

    @staticmethod
    def _default_base_url() -> str:
        return "https://darwin.v7labs.com/"

    @staticmethod
    def _default_api_endpoint() -> str:
        return DarwinConfig._default_base_url() + "api/"

    @staticmethod
    def from_file(path: Path) -> DarwinConfig:
        if path.suffix.lower() == ".yaml":
            data = DarwinConfig._parse_yaml(path)
            return DarwinConfig.parse_obj(data)
        else:
            return DarwinConfig.parse_file(path)

    @staticmethod
    def _parse_yaml(path: Path) -> dict:
        with open(path, encoding="utf-8") as infile:
            data = yaml.safe_load(infile)
        return data

    @staticmethod
    def from_api_key_with_defaults(api_key: str) -> DarwinConfig:
        return DarwinConfig(
            api_key=api_key,
            api_endpoint=DarwinConfig._default_api_endpoint(),
            base_url=DarwinConfig._default_base_url(),
            default_team="default",
            teams={},
            datasets_dir=DarwinConfig._default_config_path(),
        )

    class Config:
        validate_assignment = True


class Result(BaseModel):
    """Default model for a result, to be extended by other models specific to the API"""

    ...


class Client:
    """Client Object to manage and make requests to the Darwin API
    Attributes
    ----------
    url: str, url of the endpoint
    api_key: str, api key to authenticate
    team: Team, team to make requests to
    """

    def __init__(self, config: DarwinConfig, retries: Optional[Retry] = None) -> None:
        self.config = config
        self.session = requests.Session()
        if not retries:
            retries = Retry(total=3, backoff_factor=0.2, status_forcelist=[500, 502, 503, 504])
        self._setup_session(retries)
        self._mappings = {
            "get": self.session.get,
            "put": self.session.put,
            "post": self.session.post,
            "delete": self.session.delete,
            "patch": self.session.patch,
        }

    def _setup_session(self, retries: Retry) -> None:
        self.session.headers.update(self.headers)
        self.session.mount("http://", HTTPAdapter(max_retries=retries))
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

    @property
    def headers(self) -> Dict[str, str]:
        http_headers: Dict[str, str] = {"Content-Type": "application/json"}
        if self.config.api_key:
            http_headers["Authorization"] = f"ApiKey {self.config.api_key}"
        return http_headers

    @overload
    def _generic_call(self, method: Callable[[str], requests.Response], endpoint: str) -> dict:
        ...

    @overload
    def _generic_call(self, method: Callable[[str, dict], requests.Response], endpoint: str, payload: dict) -> dict:
        ...

    def _generic_call(self, method: Callable, endpoint: str, payload: Optional[dict] = None) -> JSONType:
        endpoint = self._sanitize_endpoint(endpoint)
        url = self.config.api_endpoint + endpoint
        if payload is not None:
            response = method(url, payload)
        else:
            response = method(url)

        raise_for_darwin_exception(response)
        response.raise_for_status()

        return response.json()

    def _contain_qs_and_endpoint(self, endpoint: str, query_string: Optional[QueryString] = None) -> str:
        if not query_string:
            return endpoint

        assert "?" not in endpoint
        return endpoint + str(query_string)

    def get(self, endpoint: str, query_string: Optional[QueryString] = None) -> JSONType:
        return self._generic_call(self.session.get, self._contain_qs_and_endpoint(endpoint, query_string))

    def put(self, endpoint: str, data: dict) -> JSONType:
        return self._generic_call(self.session.put, endpoint, data)

    def post(self, endpoint: str, data: dict) -> JSONType:
        return self._generic_call(self.session.post, endpoint, data)

    def delete(self, endpoint: str, query_string: Optional[QueryString] = None) -> JSONType:
        return self._generic_call(self.session.delete, self._contain_qs_and_endpoint(endpoint, query_string))

    def patch(self, endpoint: str, data: dict) -> JSONType:
        return self._generic_call(self.session.patch, endpoint, data)

    def _sanitize_endpoint(self, endpoint: str) -> str:
        return endpoint.strip().strip("/")


def raise_for_darwin_exception(response: requests.Response) -> None:
    """Raises an exception if the response.status_code matches a known darwin error

    Parameters
    ----------
    response: requests.Response, response to check
    """
    if response.status_code == 200:
        return
    if response.status_code == 401:
        raise Unauthorized(response)
    if response.status_code == 404:
        raise NotFound(response)



from typing import List, Tuple

from pydantic import parse_obj_as

from darwin.future.core.client import Client
from darwin.future.data_objects.dataset import Dataset


def list_datasets(api_client: Client) -> Tuple[List[Dataset], List[Exception]]:
    """
    Returns a list of datasets for the given team

    Parameters
    ----------
    api_client : Client
        The client to use to make the request
    team_slug : Optional[TeamSlug]
        The slug of the team to retrieve datasets for

    Returns
    -------
    Tuple[DatasetList, List[Exception]]
    """
    datasets: List[Dataset] = []
    errors: List[Exception] = []

    try:
        response = api_client.get("/datasets")
        for item in response:
            datasets.append(parse_obj_as(Dataset, item))
    except Exception as e:
        errors.append(e)

    return datasets, errors



from json import loads
from pathlib import Path
from uuid import UUID

import pytest

from darwin.future.data_objects.workflow import WFStage
from darwin.future.tests.data_objects.fixtures import test_data_path

validate_json = test_data_path / "stage.json"


def test_file_exists() -> None:
    # This is a test sanity check to make sure the file exists
    # Helps avoids headaches when debugging tests
    assert validate_json.exists()


def test_WFStage_validates_from_valid_json() -> None:
    WFStage.parse_file(validate_json)
    assert True


def test_casts_strings_to_uuids_as_needed() -> None:
    parsed_stage = WFStage.parse_file(validate_json)
    assert isinstance(parsed_stage.id, UUID)
    assert str(parsed_stage.id) == "e69d3ebe-6ab9-4159-b44f-2bf84d29bb20"


def test_raises_with_invalid_uuid() -> None:
    dict_from_json = loads(validate_json.read_text())
    dict_from_json["id"] = "not-a-uuid"

    with pytest.raises(ValueError) as excinfo:
        WFStage.parse_obj(dict_from_json)

    assert "not a valid uuid" in str(excinfo.value)
    assert str(excinfo.value).startswith("1 validation error for WFStage\nid")



import pytest
import responses
from pydantic import ValidationError
from requests import HTTPError

from darwin.future.core.client import Client, JSONType
from darwin.future.core.workflows.get_workflow import get_workflow
from darwin.future.data_objects.workflow import Workflow
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_get_workflow(base_client: Client, base_single_workflow_object: JSONType) -> None:
    # Mocking the response using responses library
    response_data = base_single_workflow_object
    workflow_id = "1"
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows/{workflow_id}",
        json=response_data,
        status=200,
    )

    # Call the function being tested
    workflow, exceptions = get_workflow(base_client, workflow_id)

    # Assertions
    assert isinstance(workflow, Workflow)
    assert not exceptions


@responses.activate
def test_get_workflow_with_team_slug(base_client: Client, base_single_workflow_object: JSONType) -> None:
    # Mocking the response using responses library
    team_slug = "team-slug"
    workflow_id = "1"

    response_data = base_single_workflow_object
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{team_slug}/workflows/{workflow_id}",
        json=response_data,
        status=200,
    )

    # Call the function being tested
    workflow, exceptions = get_workflow(base_client, workflow_id, team_slug)

    # Assertions
    assert isinstance(workflow, Workflow)
    assert not exceptions


@responses.activate
def test_get_workflows_with_invalid_response(base_client: Client) -> None:
    # Mocking the response using responses library
    # fmt: off
    NON_EXISTENT_ID = "1"
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows/{NON_EXISTENT_ID}",
        json="{}",
        status=200
    )
    # fmt: on

    # Call the function being tested
    workflow, exceptions = get_workflow(base_client, NON_EXISTENT_ID)

    assert not workflow
    assert exceptions
    assert len(exceptions) == 1
    assert isinstance(exceptions[0], ValidationError)


@responses.activate
def test_get_workflows_with_error(base_client: Client) -> None:
    # Mocking the response using responses library
    # fmt: off
    NON_EXISTENT_ID = "1"
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows/{NON_EXISTENT_ID}",
        json="{}",
        status=400
    )
    # fmt: on

    workflow, exceptions = get_workflow(base_client, NON_EXISTENT_ID)

    assert not workflow
    assert exceptions
    assert len(exceptions) == 1
    assert isinstance(exceptions[0], HTTPError)



from typing import List

import pytest
import responses
from pydantic import ValidationError
from requests import HTTPError

from darwin.future.core.client import Client, JSONType
from darwin.future.core.workflows.get_workflows import get_workflows
from darwin.future.data_objects.workflow import Workflow
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_get_workflows(base_client: Client, base_workflows_object: str) -> None:
    # Mocking the response using responses library
    response_data = base_workflows_object
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows?worker=false",
        json=response_data,
        status=200,
    )

    # Call the function being tested
    workflows = get_workflows(base_client)

    # Assertions
    assert isinstance(workflows, List)
    assert len(workflows) == 3
    assert all(isinstance(workflow, Workflow) for workflow in workflows)


@responses.activate
def test_get_workflows_with_team_slug(base_client: Client, base_workflows_object: JSONType) -> None:
    # Mocking the response using responses library
    team_slug = "team-slug"
    response_data = base_workflows_object
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{team_slug}/workflows?worker=false",
        json=response_data,
        status=200,
    )

    # Call the function being tested
    workflows = get_workflows(base_client, team_slug=team_slug)

    # Assertions
    assert isinstance(workflows, List)
    assert len(workflows) == len(response_data)
    assert all(isinstance(workflow, Workflow) for workflow in workflows)


@responses.activate
def test_get_workflows_with_invalid_response(base_client: Client) -> None:
    # Mocking the response using responses library
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows?worker=false",
        json="{}",
        status=200,
    )

    # Call the function being tested
    with pytest.raises(ValidationError):
        get_workflows(base_client)


@responses.activate
def test_get_workflows_with_error(base_client: Client) -> None:
    # Mocking the response using responses library
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows?worker=false",
        json="",
        status=400,
    )

    # Call the function being tested
    with pytest.raises(HTTPError):
        get_workflows(base_client)



from typing import Type


def assert_is(conditional: bool, message: str, exception_factory: Type[BaseException] = AssertionError) -> None:
    if not conditional:
        raise exception_factory(message)



from typing import Optional
from uuid import UUID

from darwin.future.core.client import Client
from darwin.future.data_objects.workflow import WFStage
from darwin.future.meta.objects.base import MetaBase


class StageMeta(MetaBase[WFStage]):
    """_summary_

    Args:
        MetaBase (_type_): _description_
    """

    def __init__(
        self,
        client: Client,
        item: Optional[WFStage] = None,
        workflow_id: Optional[UUID] = None,
    ) -> None:
        self._workflow_id = workflow_id
        super().__init__(client, item)



from typing import List, Optional, Tuple

from pydantic import parse_obj_as

from darwin.future.core.client import Client
from darwin.future.data_objects.workflow import Workflow


def get_workflow(
    client: Client, workflow_id: str, team_slug: Optional[str] = None
) -> Tuple[Optional[Workflow], List[Exception]]:
    workflow: Optional[Workflow] = None
    exceptions: List[Exception] = []

    try:
        team_slug = team_slug or client.config.default_team
        response = client.get(f"/v2/teams/{team_slug}/workflows/{workflow_id}")

        workflow = parse_obj_as(Workflow, response)
    except Exception as e:
        exceptions.append(e)

    return workflow, exceptions



from typing import List

import pytest
import responses
from pydantic import ValidationError
from requests import HTTPError

from darwin.future.core.client import Client, JSONType
from darwin.future.core.workflows.list_workflows import list_workflows
from darwin.future.data_objects.workflow import Workflow
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_list_workflows(base_client: Client, base_workflows_object: str) -> None:
    # Mocking the response using responses library
    response_data = base_workflows_object
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows?worker=false",
        json=response_data,
        status=200,
    )

    # Call the function being tested
    workflows, exceptions = list_workflows(base_client)

    # Assertions
    assert isinstance(workflows, List)
    assert len(workflows) == 3
    assert all(isinstance(workflow, Workflow) for workflow in workflows)

    assert not exceptions


@responses.activate
def test_list_workflows_with_team_slug(base_client: Client, base_workflows_object: JSONType) -> None:
    # Mocking the response using responses library
    team_slug = "team-slug"
    response_data = base_workflows_object
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{team_slug}/workflows?worker=false",
        json=response_data,
        status=200,
    )

    # Call the function being tested
    workflows, exceptions = list_workflows(base_client, team_slug=team_slug)

    # Assertions
    assert isinstance(workflows, List)
    assert len(workflows) == len(response_data)
    assert all(isinstance(workflow, Workflow) for workflow in workflows)

    assert not exceptions


@responses.activate
def test_list_workflows_with_invalid_response(base_client: Client) -> None:
    # Mocking the response using responses library
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows?worker=false",
        json="INVALIDRESPONSE",
        status=200,
    )

    # Call the function being tested
    workflows, exceptions = list_workflows(base_client)

    assert isinstance(exceptions, List)
    assert len(exceptions) == 1
    assert isinstance(exceptions[0], ValidationError)

    assert not workflows


@responses.activate
def test_list_workflows_with_error(base_client: Client) -> None:
    # Mocking the response using responses library
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows?worker=false",
        json="",
        status=400,
    )

    # Call the function being tested
    workflows, exceptions = list_workflows(base_client)

    assert isinstance(exceptions, List)
    assert len(exceptions) == 1
    assert isinstance(exceptions[0], HTTPError)

    assert not workflows



import responses
from pydantic import ValidationError
from pytest import raises
from requests import HTTPError

from darwin.future.core.client import Client
from darwin.future.core.datasets.get_dataset import get_dataset
from darwin.future.data_objects.dataset import Dataset
from darwin.future.tests.core.fixtures import *

from .fixtures import *


def test_it_gets_a_dataset(base_client: Client, basic_dataset: Dataset) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint + "datasets",
            json=basic_dataset,
            status=200,
        )

        dataset = get_dataset(base_client, "test-dataset")

        assert dataset.name == "test-dataset"
        assert dataset.slug == "1337"


def test_it_raises_an_error_on_http_error(base_client: Client) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint + "datasets",
            json={},
            status=400,
        )
        with raises(HTTPError):
            get_dataset(base_client, "test-dataset")
            get_dataset(base_client, "test-dataset")


def test_it_raises_an_error_on_pydantic_error(sad_client_pydantic: Client) -> None:
    with raises(ValidationError):
        get_dataset(sad_client_pydantic, "test-dataset")



from typing import List, Optional

from darwin.future.core.client import Client
from darwin.future.data_objects.team import Team, get_team
from darwin.future.helpers.assertion import assert_is
from darwin.future.meta.objects.base import MetaBase
from darwin.future.meta.queries.team_member import TeamMemberQuery


class TeamMeta(MetaBase[Team]):
    """Team Meta object. Facilitates the creation of Query objects, lazy loading of sub fields like members
    unlike other MetaBase objects, does not extend the __next__ function because it is not iterable. This is because
    Team is linked to api key and only one team can be returned, but stores a list of teams for consistency. This
    does mean however that to access the underlying team object, you must access the first element of the list
    team = client.team[0]

    Args:
        MetaBase (Team): Generic MetaBase object expanded by Team core object return type

    Returns:
        _type_: TeamMeta
    """

    def __init__(self, client: Client, team: Optional[Team] = None) -> None:
        team = team or get_team(client)
        super().__init__(client, team)

    @property
    def members(self) -> TeamMemberQuery:
        return TeamMemberQuery(self.client)

    # @property
    # def workflows(self) -> WorkflowQuery:
    #     return WorkflowQuery(self.client)



from uuid import UUID

from darwin.future.data_objects.workflow import Workflow
from darwin.future.meta.objects.base import MetaBase
from darwin.future.meta.queries.stage import StageQuery


class WorkflowMeta(MetaBase[Workflow]):
    @property
    def stages(self) -> StageQuery:
        if self._item is None:
            raise ValueError("WorkflowMeta has no item")
        meta_params = {"workflow_id": self._item.id}
        return StageQuery(self.client, meta_params=meta_params)

    @property
    def id(self) -> UUID:
        if self._item is None:
            raise ValueError("WorkflowMeta has no item")
        return self._item.id

    @property
    def name(self) -> str:
        if self._item is None:
            raise ValueError("WorkflowMeta has no item")
        return self._item.name


import responses
from pytest import raises
from requests import HTTPError

from darwin.future.core.client import Client
from darwin.future.core.datasets.remove_dataset import remove_dataset
from darwin.future.tests.core.fixtures import *

from .fixtures import *


def test_it_deletes_a_dataset(base_client: Client) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.PUT,
            base_client.config.api_endpoint + "datasets/1337/archive",
            json={
                "id": 1337,
            },
            status=200,
        )

        output = remove_dataset(base_client, 1337)

        assert output == 1337


def test_it_throws_http_errors_returned_by_the_client(base_client: Client) -> None:
    with raises(HTTPError):
        with responses.RequestsMock() as rsps:
            rsps.add(
                rsps.PUT,
                base_client.config.api_endpoint + "datasets/test-dataset/archive",
                json={
                    "affected_item_count": 1,
                },
                status=400,
            )

            remove_dataset(base_client, "test-dataset")  # type: ignore



from collections import namedtuple
from datetime import datetime, timezone

import responses

from darwin.future.core.client import Client
from darwin.future.core.types.query import Modifier
from darwin.future.data_objects.workflow import Workflow
from darwin.future.meta.objects.workflow import WorkflowMeta
from darwin.future.meta.queries.workflow import WorkflowQuery
from darwin.future.tests.core.fixtures import *

WORKFLOW_1 = "6dca86a3-48fb-40cc-8594-88310f5f1fdf"
WORKFLOW_2 = "e34fe935-4a1c-4231-bb55-454e2ac7673f"
WORKFLOW_3 = "45cf0abe-58a2-4878-b171-4fb5421a1c39"


def workflows_query_endpoint(team: str) -> str:
    return f"v2/teams/{team}/workflows?worker=false"


@responses.activate
def test_workflowquery_collects_basic(base_client: Client, base_filterable_workflows: dict) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    query = WorkflowQuery(base_client, [])
    workflows = query.collect()

    assert len(workflows) == 3
    assert all([isinstance(workflow, WorkflowMeta) for workflow in workflows])


@responses.activate
def test_workflowquery_filters_uuid(base_client: Client, base_filterable_workflows: dict) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    query = WorkflowQuery(base_client, []).where(
        {
            "name": "id",
            "param": "6dca86a3-48fb-40cc-8594-88310f5f1fdf",
        }
    )
    workflows = query.collect()

    assert len(workflows) == 1
    assert str(workflows[0].id) == WORKFLOW_1


@responses.activate
def test_workflowquery_filters_inserted_at(base_client: Client, base_filterable_workflows: dict) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    start = "2021-06-01T15:00:00.000+00:00"
    end = "2021-06-04T15:00:00.000+00:00"

    query = (
        WorkflowQuery(base_client, [])
        .where(
            {
                "name": "inserted_at_start",
                "param": start,
            }
        )
        .where(
            {
                "name": "inserted_at_end",
                "param": end,
            }
        )
    )
    workflows = query.collect()

    assert len(workflows) == 2
    ids = [str(workflow.id) for workflow in workflows]
    assert WORKFLOW_1 in ids
    assert WORKFLOW_2 in ids


@responses.activate
def test_workflowquery_filters_updated_at(base_client: Client, base_filterable_workflows: dict) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    start = "2021-06-04T15:00:00.000+00:00"
    end = "2021-06-06T15:00:00.000+00:00"

    query = (
        WorkflowQuery(base_client, [])
        .where(
            {
                "name": "updated_at_start",
                "param": start,
            }
        )
        .where(
            {
                "name": "updated_at_end",
                "param": end,
            }
        )
    )
    workflows = query.collect()

    assert len(workflows) == 2
    ids = [str(workflow.id) for workflow in workflows]
    assert WORKFLOW_1 in ids
    assert WORKFLOW_2 in ids


@responses.activate
def test_workflowquery_filters_dataset_id(base_client: Client, base_filterable_workflows: dict) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    query = WorkflowQuery(base_client, []).where(
        {
            "name": "dataset_id",
            "param": "1",
        }
    )
    workflows = query.collect()

    assert len(workflows) == 1
    assert str(workflows[0].id) == WORKFLOW_1


@responses.activate
def test_workflowquery_filters_dataset_id_multiple_ids(base_client: Client, base_filterable_workflows: dict) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    query = WorkflowQuery(base_client, []).where(
        {
            "name": "dataset_id",
            "param": "1,2",
        }
    )
    workflows = query.collect()

    assert len(workflows) == 2
    assert str(workflows[0].id) == WORKFLOW_1
    assert str(workflows[1].id) == WORKFLOW_2


@responses.activate
def test_workflowquery_filters_dataset_name(base_client: Client, base_filterable_workflows: dict) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    query = WorkflowQuery(base_client, []).where(
        {
            "name": "dataset_name",
            "param": "test-dataset-1",
        }
    )
    workflows = query.collect()

    assert len(workflows) == 1
    assert str(workflows[0].id) == WORKFLOW_1


@responses.activate
def test_workflowquery_filters_dataset_name_mutliple_names(
    base_client: Client, base_filterable_workflows: dict
) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    query = WorkflowQuery(base_client, []).where(
        {
            "name": "dataset_name",
            "param": "test-dataset-1,test-dataset-2",
        }
    )
    workflows = query.collect()

    assert len(workflows) == 2
    assert str(workflows[0].id) == WORKFLOW_1
    assert str(workflows[1].id) == WORKFLOW_2


@responses.activate
def test_workflowquery_filters_stages(base_client: Client, base_filterable_workflows: dict) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    query = WorkflowQuery(base_client, []).where(
        {
            "name": "has_stages",
            "param": "5445adcb-193d-4f76-adb0-0c6d5f5e4c04",
        }
    )
    workflows = query.collect()

    assert len(workflows) == 1
    assert str(workflows[0].name) == "test-workflow-3"


@responses.activate
def test_workflowquery_filters_stages_multiple(base_client: Client, base_filterable_workflows: dict) -> None:
    endpoint = base_client.config.api_endpoint + workflows_query_endpoint(base_client.config.default_team)
    responses.add(responses.GET, endpoint, json=base_filterable_workflows)

    query = WorkflowQuery(base_client, []).where(
        {
            "name": "has_stages",
            "param": "5445adcb-193d-4f76-adb0-0c6d5f5e4c04,53d2c997-6bb0-4766-803c-3c8d1fb21072",
        }
    )
    workflows = query.collect()

    assert len(workflows) == 2
    workflow_names = [workflow.name for workflow in workflows]

    assert "test-workflow-3" in workflow_names
    assert "test-workflow-1" in workflow_names


# Test static methods
def test__date_compare() -> None:
    date1 = datetime(1970, 1, 1, 0, 0, 0, 1, timezone.utc)
    date1_copy = datetime(1970, 1, 1, 0, 0, 0, 1, timezone.utc)
    date2 = datetime(1970, 1, 1, 0, 0, 0, 0, timezone.utc)

    assert WorkflowQuery._date_compare(date1, date2)
    assert WorkflowQuery._date_compare(date1, date1_copy)
    assert not WorkflowQuery._date_compare(date2, date1)


def test__stages_contains() -> None:
    FakeStage = namedtuple("FakeStage", ["id"])
    stages = [
        FakeStage("1"),
        FakeStage("2"),
        FakeStage("3"),
    ]

    assert WorkflowQuery._stages_contains(stages, "1")  # type: ignore
    assert WorkflowQuery._stages_contains(stages, "2")  # type: ignore
    assert WorkflowQuery._stages_contains(stages, "3")  # type: ignore
    assert not WorkflowQuery._stages_contains(stages, "4")  # type: ignore



from pydantic import parse_obj_as

from darwin.future.core.client import Client
from darwin.future.core.types.common import QueryString
from darwin.future.data_objects.dataset import Dataset


def get_dataset(api_client: Client, dataset_id: str) -> Dataset:
    """
    Returns a list of datasets for the given team

    Parameters
    ----------
    api_client : Client
        The client to use to make the request
    dataset_id : str
        The id of the dataset to retrieve

    Returns
    -------
    Dataset

    Raises
    ------
    HTTPError
        Any errors that occurred while making the request
    ValidationError
        Any errors that occurred while parsing the response
    """

    response = api_client.get("/datasets", QueryString({"id": str(dataset_id)}))

    return parse_obj_as(Dataset, response)



from typing import List, Optional, Tuple, Union

from darwin.future.core.client import Client
from darwin.future.core.datasets.create_dataset import create_dataset
from darwin.future.core.datasets.get_dataset import get_dataset
from darwin.future.core.datasets.list_datasets import list_datasets
from darwin.future.core.datasets.remove_dataset import remove_dataset
from darwin.future.data_objects.dataset import Dataset
from darwin.future.helpers.assertion import assert_is
from darwin.future.meta.objects.base import MetaBase


class DatasetMeta(MetaBase[Dataset]):
    """Dataset Meta object. Facilitates the creation of Query objects, lazy loading of sub fields

    Args:
        MetaBase (Dataset): Generic MetaBase object expanded by Dataset core object return type

    Returns:
        _type_: DatasetMeta
    """

    def get_dataset_by_id(self) -> Dataset:
        # TODO: implement
        raise NotImplementedError()

    def create_dataset(self, slug: str) -> Tuple[Optional[List[Exception]], Optional[Dataset]]:
        """
        Creates a new dataset for the given team

        Parameters
        ----------
        slug: str [a-b0-9-_]
            The slug of the dataset to create

        Returns
        -------
        Tuple[Optional[List[Exception]], Optional[Dataset]]
            A tuple containing a list of exceptions and the dataset created

        """
        exceptions = []
        dataset: Optional[Dataset] = None

        try:
            self._validate_slug(slug)
            dataset = create_dataset(self.client, slug)
        except Exception as e:
            exceptions.append(e)

        return exceptions or None, dataset

    def update_dataset(self) -> Dataset:
        # TODO: implement in IO-1018
        raise NotImplementedError()

    def delete_dataset(self, dataset_id: Union[int, str]) -> Tuple[Optional[List[Exception]], int]:
        """
        Deletes a dataset by id or slug

        Parameters
        ----------
        dataset_id: Union[int, str]
            The id or slug of the dataset to delete

        Returns
        -------
        Tuple[Optional[List[Exception]], int]
            A tuple containing a list of exceptions and the number of datasets deleted
        """
        exceptions = []
        dataset_deleted = -1

        try:
            if isinstance(dataset_id, str):
                dataset_deleted = self._delete_by_slug(self.client, dataset_id)
            else:
                dataset_deleted = self._delete_by_id(self.client, dataset_id)

        except Exception as e:
            exceptions.append(e)

        return exceptions or None, dataset_deleted

    @staticmethod
    def _delete_by_slug(client: Client, slug: str) -> int:
        """
        (internal) Deletes a dataset by slug

        Parameters
        ----------
        client: MetaClient
            The client to use to make the request

        slug: str
            The slug of the dataset to delete

        Returns
        -------
        int
            The dataset deleted
        """
        assert_is(isinstance(client, Client), "client must be a Core Client")
        assert_is(isinstance(slug, str), "slug must be a string")

        dataset = get_dataset(client, slug)
        if dataset and dataset.id:
            dataset_deleted = remove_dataset(client, dataset.id)
        else:
            raise Exception(f"Dataset with slug {slug} not found")

        return dataset_deleted

    @staticmethod
    def _delete_by_id(client: Client, dataset_id: int) -> int:
        """
        (internal) Deletes a dataset by id

        Parameters
        ----------
        client: Client
            The client to use to make the request

        dataset_id: int
            The id of the dataset to delete

        Returns
        -------
        int
            The dataset deleted
        """
        assert_is(isinstance(client, Client), "client must be a Client")
        assert_is(isinstance(dataset_id, int), "dataset_id must be an integer")

        dataset_deleted = remove_dataset(client, dataset_id)
        return dataset_deleted

    @staticmethod
    def _validate_slug(slug: str) -> None:
        """
        Validates a slug

        Parameters
        ----------
        slug: str
            The slug to validate

        Raises
        ------
        AssertionError
        """
        slug_copy = str(slug).lower().strip()
        assert_is(isinstance(slug_copy, str), "slug must be a string")
        assert_is(len(slug_copy) > 0, "slug must not be empty")

        VALID_SLUG_CHARS = "abcdefghijklmnopqrstuvwxyz0123456789-_"
        assert_is(all(c in VALID_SLUG_CHARS for c in slug_copy), "slug must only contain valid characters")



from typing import List, Optional

from darwin.future.core.client import Client
from darwin.future.data_objects.team import TeamMember, get_team_members
from darwin.future.data_objects.team_member_role import TeamMemberRole
from darwin.future.meta.objects.base import MetaBase


class TeamMemberMeta(MetaBase[TeamMember]):
    @property
    def role(self) -> TeamMemberRole:
        if self._item is None:
            raise ValueError("TeamMemberMeta has no item")
        return self._item.role



import string
from unittest.mock import Mock, patch

from pytest import fixture, mark, raises
from responses import RequestsMock

from darwin.future.core.client import DarwinConfig
from darwin.future.meta.client import MetaClient
from darwin.future.meta.objects.dataset import DatasetMeta
from darwin.future.tests.core.fixtures import *


@fixture
def _delete_by_slug_mock():  # type: ignore
    with patch.object(DatasetMeta, "_delete_by_slug") as mock:
        yield mock


@fixture
def _delete_by_id_mock():  # type: ignore
    with patch.object(DatasetMeta, "_delete_by_id") as mock:
        yield mock


# `datasets` tests
# TODO datasets tests

# `get_dataset_by_id` tests
# TODO get_dataset_by_id tests


# `create_dataset` tests
def test_create_dataset_returns_exceptions_thrown(base_config: DarwinConfig) -> None:
    valid_client = MetaClient(base_config)
    valid_slug = "test_dataset"

    base_url = base_config.base_url + "api/datasets"

    with RequestsMock() as rsps:
        rsps.add(rsps.POST, base_url, status=500)
        dataset_meta = DatasetMeta(valid_client)

        exceptions, dataset_created = dataset_meta.create_dataset(valid_slug)

        assert exceptions is not None
        assert "500 Server Error" in str(exceptions[0])
        assert dataset_created is None


def test_create_dataset_returns_dataset_created_if_dataset_created(base_config: DarwinConfig) -> None:
    valid_client = MetaClient(base_config)
    valid_slug = "test_dataset"

    base_url = base_config.base_url + "api/datasets"

    with RequestsMock() as rsps:
        rsps.add(
            rsps.POST,
            base_url,
            json={"id": 1, "name": "Test Dataset", "slug": "test_dataset"},
            status=201,
        )
        dataset_meta = DatasetMeta(valid_client)

        exceptions, dataset_created = dataset_meta.create_dataset(valid_slug)

        assert exceptions is None
        assert dataset_created is not None
        assert dataset_created.id == 1
        assert dataset_created.name == "test dataset"
        assert dataset_created.slug == "test_dataset"


# `update_dataset` tests
# TODO update_dataset tests


# `delete_dataset` tests
def test_delete_dataset_returns_exceptions_thrown(
    base_config: DarwinConfig, _delete_by_id_mock: Mock, _delete_by_slug_mock: Mock
) -> None:
    _delete_by_slug_mock.side_effect = Exception("test exception")

    client = MetaClient(base_config)
    dataset_meta = DatasetMeta(client)

    exceptions, dataset_deleted = dataset_meta.delete_dataset("test_dataset")

    assert exceptions is not None
    assert str(exceptions[0]) == "test exception"
    assert dataset_deleted == -1

    assert _delete_by_slug_mock.call_count == 1
    assert _delete_by_id_mock.call_count == 0


def test_delete_dataset_calls_delete_by_slug_as_appropriate(
    base_config: DarwinConfig, _delete_by_id_mock: Mock, _delete_by_slug_mock: Mock
) -> None:
    client = MetaClient(base_config)
    dataset_meta = DatasetMeta(client)

    exceptions, _ = dataset_meta.delete_dataset("test_dataset")

    assert exceptions is None
    assert _delete_by_slug_mock.call_count == 1
    assert _delete_by_id_mock.call_count == 0


def test_delete_dataset_calls_delete_by_id_as_appropriate(
    base_config: DarwinConfig, _delete_by_id_mock: Mock, _delete_by_slug_mock: Mock
) -> None:
    client = MetaClient(base_config)
    dataset_meta = DatasetMeta(client)

    exceptions, _ = dataset_meta.delete_dataset(1)

    assert exceptions is None
    assert _delete_by_slug_mock.call_count == 0
    assert _delete_by_id_mock.call_count == 1


# Test `_delete_by_slug`
def test_delete_by_slug_raises_exception_if_not_passed_str_and_client(base_config: DarwinConfig) -> None:
    valid_client = MetaClient(base_config)
    valid_slug = "test_dataset"
    invalid_client = "client"
    invalid_slug = 1

    with raises(AssertionError):
        DatasetMeta._delete_by_slug(valid_client, invalid_slug)  # type: ignore

    with raises(AssertionError):
        DatasetMeta._delete_by_slug(invalid_client, valid_slug)  # type: ignore


def test_delete_by_slug__returns_dataset_deleted_if_dataset_found(base_config: DarwinConfig) -> None:
    valid_client = MetaClient(base_config)
    valid_slug = "test_dataset"

    base_url = base_config.base_url + "api/datasets"

    with RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_url + "?id=test_dataset",
            json={"id": 1, "name": "Test Dataset", "slug": "test_dataset"},
            status=200,
        )
        rsps.add(
            rsps.PUT,
            base_url + "/1/archive",
            json={"id": 1, "name": "Test Dataset", "slug": "test_dataset"},
            status=200,
        )
        dataset_deleted = DatasetMeta._delete_by_slug(valid_client, valid_slug)

        assert dataset_deleted == 1


# Test `_delete_by_id`
def test_delete_by_id_raises_exception_if_not_passed_int_and_client(base_config: DarwinConfig) -> None:
    valid_client = MetaClient(base_config)
    valid_id = 1
    invalid_client = "client"
    invalid_id = "1"

    with raises(AssertionError):
        DatasetMeta._delete_by_id(valid_client, invalid_id)  # type: ignore

    with raises(AssertionError):
        DatasetMeta._delete_by_id(invalid_client, valid_id)  # type: ignore


def test_delete_by_id_returns_dataset_deleted_if_dataset_found(base_config: DarwinConfig) -> None:
    valid_client = MetaClient(base_config)
    valid_id = 1

    base_url = base_config.base_url + "api/datasets"

    with RequestsMock() as rsps:
        rsps.add(
            rsps.PUT,
            base_url + "/1/archive",
            json={"id": 1, "name": "Test Dataset", "slug": "test_dataset"},
            status=200,
        )
        dataset_deleted = DatasetMeta._delete_by_id(valid_client, valid_id)

        assert dataset_deleted == 1


@mark.parametrize(
    "invalid_slug",
    ["", " ", "test dataset", *[f"dataset_{c}" for c in string.punctuation if c not in ["-", "_", "."]]],
)
def test_validate_slugh_raises_exception_if_passed_invalid_inputs(invalid_slug: str) -> None:
    with raises(AssertionError):
        DatasetMeta._validate_slug(invalid_slug)


def test_validate_slug_returns_none_if_passed_valid_slug() -> None:
    valid_slug = "test-dataset"

    assert DatasetMeta._validate_slug(valid_slug) is None



from darwin.future.exceptions.base import DarwinException


class DatasetException(DarwinException):
    """Base class for all dataset exceptions."""

    ...


class DatasetNotFound(DatasetException):
    """Raised when the dataset endpoint returns a malformed response."""

    ...



from typing import List

import pytest
import responses

from darwin.future.core.client import Client
from darwin.future.data_objects.team import TeamMember
from darwin.future.data_objects.team_member_role import TeamMemberRole
from darwin.future.meta.objects.team_member import TeamMemberMeta
from darwin.future.meta.queries.team_member import TeamMemberQuery
from darwin.future.tests.core.fixtures import *


def test_team_member_collects_basic(base_client: Client, base_team_members_json: List[dict]) -> None:
    query = TeamMemberQuery(base_client)
    with responses.RequestsMock() as rsps:
        endpoint = base_client.config.api_endpoint + "memberships"
        rsps.add(responses.GET, endpoint, json=base_team_members_json)
        members = query.collect()
        assert len(members) == len(TeamMemberRole)
        assert isinstance(members[0], TeamMemberMeta)


def test_team_member_only_passes_back_correct(base_client: Client, base_team_member_json: dict) -> None:
    query = TeamMemberQuery(base_client)
    with responses.RequestsMock() as rsps:
        endpoint = base_client.config.api_endpoint + "memberships"
        rsps.add(responses.GET, endpoint, json=[base_team_member_json, {}])
        members = query.collect()
        assert len(members) == 1
        assert isinstance(members[0], TeamMemberMeta)


@pytest.mark.parametrize("role", [role for role in TeamMemberRole])
def test_team_member_filters_role(
    role: TeamMemberRole, base_client: Client, base_team_members_json: List[dict]
) -> None:
    with responses.RequestsMock() as rsps:
        # Test equal
        query = TeamMemberQuery(base_client).where({"name": "role", "param": role.value})
        endpoint = base_client.config.api_endpoint + "memberships"
        rsps.add(responses.GET, endpoint, json=base_team_members_json)
        members = query.collect()
        assert len(members) == 1
        assert members[0]._item is not None
        assert members[0]._item.role == role

        # Test not equal
        rsps.reset()
        query = TeamMemberQuery(base_client).where({"name": "role", "param": role.value, "modifier": "!="})
        rsps.add(responses.GET, endpoint, json=base_team_members_json)
        members = query.collect()
        assert len(members) == len(TeamMemberRole) - 1
        for member in members:
            assert member._item is not None
            assert member._item.role != role


def test_team_member_filters_general(base_client: Client, base_team_members_json: List[dict]) -> None:
    for idx in range(len(base_team_members_json)):
        base_team_members_json[idx]["id"] = idx + 1

    with responses.RequestsMock() as rsps:
        query = TeamMemberQuery(base_client).where({"name": "id", "param": 1})
        endpoint = base_client.config.api_endpoint + "memberships"
        rsps.add(responses.GET, endpoint, json=base_team_members_json)
        members = query.collect()
        assert len(members) == 1
        assert members[0]._item is not None
        assert members[0]._item.id == 1

        # Test chained
        rsps.reset()

        rsps.add(responses.GET, endpoint, json=base_team_members_json)

        members = (
            TeamMemberQuery(base_client)
            .where({"name": "id", "param": 1, "modifier": ">"})
            .where({"name": "id", "param": len(base_team_members_json), "modifier": "<"})
            .collect()
        )

        assert len(members) == len(base_team_members_json) - 2



import unittest

import pytest
import responses

from darwin.future.core.client import DarwinConfig
from darwin.future.data_objects.team import Team
from darwin.future.meta.client import MetaClient
from darwin.future.meta.objects.team import TeamMeta
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.meta.fixtures import *


def test_creates_from_api_key() -> None:
    with responses.RequestsMock() as rsps:
        base_api_endpoint = DarwinConfig._default_api_endpoint()
        rsps.add(responses.GET, base_api_endpoint + "users/token_info", json={"selected_team": {"slug": "test-team"}})
        client = MetaClient.from_api_key(api_key="test")
        assert client.config.default_team == "test-team"


def test_team_property(base_meta_client: MetaClient, base_team: Team, base_team_json: dict) -> None:
    client = base_meta_client
    endpoint = client.config.api_endpoint + f"teams/{client.config.default_team}"
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, endpoint, json=base_team_json)
        team = client.team
        assert isinstance(team, TeamMeta)
        assert team._item == base_team



from pytest import fixture, mark, raises
from responses import RequestsMock

from darwin.future.core.client import DarwinConfig
from darwin.future.meta.client import MetaClient
from darwin.future.meta.objects.stage import StageMeta
from darwin.future.tests.core.fixtures import *

# TODO: fill out as stage meta gets more functionality



import logging
import sys
from typing import List, Optional, Union

from rich.console import Console


def handle_exception(exception: Optional[Union[Exception, List[Exception]]]) -> None:
    """
    Handles an exception or list of exceptions by printing them to the terminal

    Parameters
    ----------
    exception : Optional[Union[Exception, List[Exception]]]
        The exception(s) to handle
    """
    IS_INTERACTIVE_SESSION = sys.stdout and sys.stdout.isatty()

    if not exception:
        exc_info = sys.exc_info()
        if exception := getattr(exc_info, "[1]", None):
            ...
        else:
            raise ValueError("No exception provided and no exception in sys.exc_info")

    if IS_INTERACTIVE_SESSION:
        console = Console()
        handler = console.print
    else:
        logger = logging.getLogger(__name__)
        handler = logger.error  # type: ignore

    if isinstance(exception, list):
        for e in exception:
            handler(e)
    else:
        handler(exception)



from __future__ import annotations

from typing import List, Optional, Tuple

from pydantic import validator

from darwin.future.core.client import Client
from darwin.future.data_objects.dataset import DatasetList
from darwin.future.data_objects.team_member_role import TeamMemberRole
from darwin.future.data_objects.validators import parse_name
from darwin.future.pydantic_base import DefaultDarwin


class TeamMember(DefaultDarwin):
    """A class to manage all the information around a team member on the darwin platform
    Attributes
    ----------
    name: str
    """

    email: str
    first_name: str
    last_name: str
    role: TeamMemberRole
    team_id: int
    id: int
    user_id: int


TeamMemberList = List[TeamMember]


class Team(DefaultDarwin):
    """A class to manage all the information around a Team on the darwin platform, including validation

    Attributes
    ----------
    slug : str
    datasets: Optional[List[Dataset]] = None
        - a list of datasets linked to the team
    members: Optional[List[Release]] = None
        - a list of members linked to a team
    Methods
    ----------
    _slug_validator: validates and auto formats the slug variable
    """

    slug: str
    id: int
    datasets: Optional[DatasetList] = None
    members: Optional[List[TeamMember]] = None
    default_role: TeamMemberRole = TeamMemberRole.USER

    # Data Validation
    _slug_validator = validator("slug", allow_reuse=True)(parse_name)

    @staticmethod
    def from_client(client: Client, team_slug: Optional[str] = None) -> Team:
        """Returns the team with the given slug from the client

        Args:
            client (Client): Core client object
            team_slug (Optional[str], optional): team slug str, Defaults to None.

        Returns:
            Team: Team object retrieved from the client with the given slug
        """
        if not team_slug:
            team_slug = client.config.default_team
        return get_team(client, team_slug)


TeamList = List[Team]


def get_team(client: Client, team_slug: Optional[str] = None) -> Team:
    """Returns the team with the given slug from the client

    Args:
        client (Client): Core client object
        team_slug (Optional[str], optional): team slug str, Defaults to None.

    Returns:
        Team: Team object retrieved from the client with the given slug
    """
    if not team_slug:
        team_slug = client.config.default_team
    response = client.get(f"/teams/{team_slug}/")
    return Team.parse_obj(response)


def get_team_members(client: Client) -> Tuple[List[TeamMember], List[Exception]]:
    """Returns a list of team members for the given client

    Args:
        client (Client): Core client object

    Returns:
        Tuple[List[TeamMember], List[Exception]]: List of team members and list of errors if any
    """
    response = client.get(f"/memberships")
    members = []
    errors = []
    for item in response:
        try:
            members.append(TeamMember.parse_obj(item))
        except Exception as e:
            errors.append(e)
    return members, errors



from datetime import datetime
from pathlib import Path
from uuid import UUID

from darwin.future.data_objects.workflow import WFDataset, WFStage, Workflow

test_data_path: Path = Path(__file__).parent / "data"
validate_json = test_data_path / "workflow.json"


def test_file_exists() -> None:
    # This is a test sanity check to make sure the file exists
    # Helps avoids headaches when debugging tests
    assert validate_json.exists()


def test_Workflow_validates_from_valid_json() -> None:
    parsed_set = Workflow.parse_file(validate_json)

    assert isinstance(parsed_set, Workflow)
    assert isinstance(parsed_set.id, UUID)
    assert isinstance(parsed_set.name, str)
    assert isinstance(parsed_set.team_id, int)

    assert isinstance(parsed_set.stages, list)
    assert all(isinstance(i, WFStage) for i in parsed_set.stages)
    assert isinstance(parsed_set.dataset, WFDataset)

    assert isinstance(parsed_set.inserted_at, datetime)
    assert isinstance(parsed_set.updated_at, datetime)

    assert isinstance(parsed_set.thumbnails, list)
    assert all(isinstance(i, str) for i in parsed_set.thumbnails)



import unittest

import pytest
from pydantic import BaseModel, ValidationError

from darwin.future.data_objects.dataset import Dataset
from darwin.future.data_objects.release import Release
from darwin.future.data_objects.team import Team
from darwin.future.tests.data_objects.fixtures import *


def test_integrated_parsing_works_with_raw(basic_combined: dict) -> None:
    team = Team.parse_obj(basic_combined)
    assert team.slug == "test-team"
    assert team.datasets is not None
    assert team.datasets[0].name == "test-dataset"
    assert team.datasets[0].releases is not None
    assert team.datasets[0].releases[0].name == "test-release"


def test_broken_obj_raises(broken_combined: dict) -> None:
    with pytest.raises(ValidationError) as e_info:
        broken = Team.parse_obj(broken_combined)


@pytest.mark.parametrize("test_object", [Team, Dataset, Release])
def test_empty_obj_raises(test_object: BaseModel) -> None:
    with pytest.raises(ValidationError) as e_info:
        broken = test_object.parse_obj({})



from datetime import datetime, timezone
from typing import List
from uuid import UUID

from darwin.exceptions import DarwinException
from darwin.future.core.client import Client
from darwin.future.core.types.query import Param, Query, QueryFilter
from darwin.future.core.workflows.list_workflows import list_workflows
from darwin.future.data_objects.workflow import WFStage, Workflow
from darwin.future.helpers.exception_handler import handle_exception
from darwin.future.meta.objects.workflow import WorkflowMeta


class WorkflowQuery(Query[WorkflowMeta]):
    """
    WorkflowQuery object with methods to manage filters, retrieve data, and execute
    filters

    Methods
    -------

    collect: Executes the query and returns the filtered data
    """

    def collect(self) -> List[WorkflowMeta]:
        workflows_core, exceptions = list_workflows(self.client)
        if exceptions:
            handle_exception(exceptions)
            raise DarwinException from exceptions[0]

        workflows = [WorkflowMeta(self.client, workflow) for workflow in workflows_core]
        if not self.filters:
            return workflows

        for filter in self.filters:
            workflows = self._execute_filters(workflows, filter)

        return workflows

    def _execute_filters(self, workflows: List[WorkflowMeta], filter: QueryFilter) -> List[WorkflowMeta]:
        if filter.name == "id":
            id_to_find = UUID(filter.param)
            return [w for w in workflows if w.id == id_to_find]

        if filter.name == "inserted_at_start":
            start_date = datetime.fromisoformat(filter.param)
            return [w for w in workflows if w._item is not None and self._date_compare(w._item.inserted_at, start_date)]

        if filter.name == "inserted_at_end":
            end_date = datetime.fromisoformat(filter.param)
            return [w for w in workflows if w._item is not None and self._date_compare(end_date, w._item.inserted_at)]

        if filter.name == "updated_at_start":
            start_date = datetime.fromisoformat(filter.param)
            return [w for w in workflows if w._item is not None and self._date_compare(w._item.updated_at, start_date)]

        if filter.name == "updated_at_end":
            end_date = datetime.fromisoformat(filter.param)
            return [w for w in workflows if w._item is not None and self._date_compare(end_date, w._item.updated_at)]

        if filter.name == "dataset_id":
            datasets_to_find_id: List[int] = [int(s) for s in filter.param.split(",")]
            return [w for w in workflows if w._item is not None and int(w._item.dataset) in datasets_to_find_id]

        if filter.name == "dataset_name":
            datasets_to_find_name: List[str] = [str(s) for s in filter.param.split(",")]
            return [w for w in workflows if w._item is not None and str(w._item.dataset) in datasets_to_find_name]

        if filter.name == "has_stages":
            stages_to_find = [s for s in filter.param.split(",")]
            return [
                w for w in workflows if w._item is not None and self._stages_contains(w._item.stages, stages_to_find)
            ]

        return self._generic_execute_filter(workflows, filter)

    @classmethod
    def _date_compare(cls, date1: datetime, date2: datetime) -> bool:
        return date1.astimezone(timezone.utc) >= date2.astimezone(timezone.utc)

    @classmethod
    def _stages_contains(cls, stages: List[WFStage], stages_to_find: List[str]) -> bool:
        stage_ids = [str(s.id) for s in stages]
        return any(stage_to_find in stage_ids for stage_to_find in stages_to_find)



from unittest.mock import MagicMock

from pydantic import BaseModel, ValidationError
from pytest import fixture
from requests import HTTPError

from darwin.future.core.client import Client


@fixture
def basic_dataset() -> dict:
    return {
        "name": "test-dataset",
        "slug": "1337",
        "id": 1,
        "releases": [],
    }


@fixture
def basic_list_of_datasets() -> list:
    return [
        {
            "name": "test-dataset",
            "slug": "1337",
            "id": 1,
            "releases": [],
        },
        {
            "name": "test-dataset-2",
            "slug": "1338",
            "id": 2,
            "releases": [],
        },
        {
            "name": "test-dataset-3",
            "slug": "1339",
            "id": 3,
            "releases": [],
        },
    ]


@fixture
def sad_http_client() -> Client:
    mock = MagicMock(Client)
    mock.post.side_effect = HTTPError("error")
    mock.get.side_effect = HTTPError("error")
    mock.delete.side_effect = HTTPError("error")

    return mock


@fixture
def happy_post_client() -> Client:
    mock_client = MagicMock(Client)
    mock_client.post.return_value = {
        "name": "test-dataset",
        "slug": "1337",
        "id": 1,
        "releases": [],
    }

    return mock_client


@fixture
def happy_get_client() -> Client:
    mock_client = MagicMock(Client)
    mock_client.get.return_value = [
        {
            "name": "test-dataset",
            "slug": "1337",
            "id": 1,
            "releases": [],
        }
    ]

    return mock_client


@fixture
def sad_client_pydantic() -> Client:
    return MagicMock(Client, side_effect=ValidationError(["error1", "error2", "error3"], model=BaseModel))



from unittest.mock import Mock, patch

import responses
from pytest import fixture, raises

from darwin.future.core.client import Client, DarwinConfig
from darwin.future.data_objects.team import Team, TeamMember
from darwin.future.meta.objects.team import TeamMeta
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.meta.objects.fixtures import *


def test_team_meta_collects_members(
    base_meta_team: TeamMeta, base_client: Client, base_team_member: TeamMember, base_team_member_json: dict
) -> None:
    with responses.RequestsMock() as rsps:
        endpoint = base_client.config.api_endpoint + "memberships"
        rsps.add(responses.GET, endpoint, json=[base_team_member_json])
        members = base_meta_team.members.collect()
        assert len(members) == 1
        assert members[0]._item == base_team_member



import json
from pathlib import Path
from typing import List

import pytest

from darwin.future.data_objects.workflow import WFStage, Workflow

test_data_path: Path = Path(__file__).parent / "workflow" / "data"
valid_stage_json = test_data_path / "stage.json"
valid_workflow_json = test_data_path / "workflow.json"


@pytest.fixture
def basic_team() -> dict:
    return {"slug": "test-team", "id": 0}


@pytest.fixture
def basic_dataset() -> dict:
    return {"name": "test-dataset", "slug": "test-dataset"}


@pytest.fixture
def basic_release() -> dict:
    return {"name": "test-release"}


@pytest.fixture
def basic_combined(basic_team: dict, basic_dataset: dict, basic_release: dict) -> dict:
    combined = basic_team
    combined["datasets"] = [basic_dataset]
    combined["datasets"][0]["releases"] = [basic_release]
    return combined


@pytest.fixture
def broken_combined(basic_combined: dict) -> dict:
    del basic_combined["datasets"][0]["name"]
    return basic_combined



from __future__ import annotations

from pathlib import Path
from typing import Optional

from requests.adapters import Retry

from darwin.future.core.client import Client, DarwinConfig
from darwin.future.meta.objects.team import TeamMeta


class MetaClient(Client):
    def __init__(self, config: DarwinConfig, retries: Optional[Retry] = None) -> None:
        self._team: Optional[TeamMeta] = None
        super().__init__(config, retries=retries)

    @classmethod
    def local(cls) -> MetaClient:
        config = DarwinConfig.local()
        return cls(config)

    @classmethod
    def from_api_key(cls, api_key: str, datasets_dir: Optional[Path] = None) -> MetaClient:
        config = DarwinConfig.from_api_key_with_defaults(api_key=api_key)
        client = Client(config)  # create a temporary client to get the default team
        token_info = client.get("/users/token_info")
        assert isinstance(token_info, dict)
        default_team: str = token_info["selected_team"]["slug"]
        config.default_team = default_team
        if datasets_dir:
            config.datasets_dir = datasets_dir
        return cls(config)

    @property
    def team(self) -> TeamMeta:
        if self._team is None:
            self._team = TeamMeta(self)
        return self._team



from typing import List, Optional, Tuple

from pydantic import ValidationError

from darwin.future.core.client import Client
from darwin.future.data_objects.workflow import Workflow, WorkflowListValidator


def list_workflows(client: Client, team_slug: Optional[str] = None) -> Tuple[List[Workflow], List[Exception]]:
    """
    Returns a list of workflows for the given team

    Parameters
    ----------
    client : Client
        The client to use to make the request
    team_slug : Optional[str]
        The slug of the team to retrieve workflows for

    Returns
    -------
    Tuple[List[Workflow], List[Exception]]
    """
    exceptions: List[Exception] = []
    workflows: List[Workflow] = []

    try:
        team_slug = team_slug or client.config.default_team
        response = client.get(f"/v2/teams/{team_slug}/workflows?worker=false")
        list_of_workflows = WorkflowListValidator(list=response)  # type: ignore
        workflows = [Workflow.parse_obj(workflow) for workflow in list_of_workflows.list]
    except Exception as e:
        exceptions.append(e)

    return workflows, exceptions



from __future__ import annotations

from typing import Generic, List, Optional, TypeVar

from darwin.future.core.client import Client
from darwin.future.pydantic_base import DefaultDarwin

R = TypeVar("R", bound=DefaultDarwin)


class MetaBase(Generic[R]):
    _item: Optional[R]
    client: Client

    def __init__(self, client: Client, item: Optional[R] = None) -> None:
        self.client = client
        self._item = item

    def __str__(self) -> str:
        if self._item is None:
            raise ValueError("MetaBase has no item")
        return str(self._item)



from __future__ import annotations

from typing import List

from darwin.future.core.types.query import Param, Query, QueryFilter
from darwin.future.data_objects.team import get_team_members
from darwin.future.meta.objects.team_member import TeamMemberMeta


class TeamMemberQuery(Query[TeamMemberMeta]):
    """TeamMemberQuery object with methods to manage filters, retrieve data, and execute filters
    Methods:
    collect: Executes the query and returns the filtered data
    _execute_filter: Executes a filter on a list of objects
    """

    def collect(self) -> List[TeamMemberMeta]:
        members, exceptions = get_team_members(self.client)
        members_meta = [TeamMemberMeta(self.client, member) for member in members]
        if exceptions:
            # TODO: print and or raise exceptions, tbd how we want to handle this
            pass
        if not self.filters:
            self.filters = []
        for filter in self.filters:
            members_meta = self._execute_filter(members_meta, filter)

        return members_meta

    def _execute_filter(self, members: List[TeamMemberMeta], filter: QueryFilter) -> List[TeamMemberMeta]:
        """Executes filtering on the local list of members, applying special logic for role filtering
        otherwise calls the parent method for general filtering on the values of the members

        Parameters
        ----------
        members : List[TeamMember]
        filter : QueryFilter

        Returns
        -------
        List[TeamMember]: Filtered subset of members
        """
        if filter.name == "role":
            return [m for m in members if m._item is not None and filter.filter_attr(m._item.role.value)]
        else:
            return super()._generic_execute_filter(members, filter)



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    overload,
)

from darwin.future.core.client import Client
from darwin.future.meta.objects.base import MetaBase
from darwin.future.pydantic_base import DefaultDarwin

T = TypeVar("T", bound=MetaBase)
R = TypeVar("R", bound=DefaultDarwin)
Param = Dict[str, Any]  # type: ignore


class Modifier(Enum):
    GREATER_THAN = ">"
    LESS_THAN = "<"
    GREATER_EQUAL = ">="
    LESS_EQUAL = "<="
    NOT_EQUAL = "!="
    CONTAINS = "contains"


class QueryFilter(DefaultDarwin):
    """Basic query filter with a name and a parameter

    Attributes
    ----------
    name: str
    param: str
    modifier: Optional[Modifiers]: Optional modifier to apply to the filter
    """

    name: str
    param: str
    modifier: Optional[Modifier] = None

    def filter_attr(self, attr: Any) -> bool:  # type: ignore
        caster: Callable[[str], Any] = type(attr)  # type: ignore
        param = caster(self.param)  # attempt to cast the param to the type of the attribute
        if self.modifier is None:
            return attr == param
        elif self.modifier == Modifier.GREATER_EQUAL:
            return attr >= param
        elif self.modifier == Modifier.GREATER_THAN:
            return attr > param
        elif self.modifier == Modifier.LESS_EQUAL:
            return attr <= param
        elif self.modifier == Modifier.LESS_THAN:
            return attr < param
        elif self.modifier == Modifier.NOT_EQUAL:
            return attr != param
        elif self.modifier == Modifier.CONTAINS:
            return param in attr
        else:
            raise ValueError(f"Unknown modifier {self.modifier}")

    @classmethod
    def _from_dict(cls, d: Dict[str, Any]) -> QueryFilter:  # type: ignore
        if "name" not in d or "param" not in d:
            raise ValueError(f"args must be a QueryFilter or a dict with 'name' and 'param' keys, got {d}")
        modifier = Modifier(d["modifier"]) if "modifier" in d else None
        return QueryFilter(name=d["name"], param=str(d["param"]), modifier=modifier)

    @classmethod
    def _from_args(cls, *args: object, **kwargs: str) -> List[QueryFilter]:
        filters = []
        for arg in args:
            filters.append(cls._from_arg(arg))
        for key, value in kwargs.items():
            filters.append(cls._from_kwarg(key, value))
        return filters

    @classmethod
    def _from_arg(cls, arg: object) -> QueryFilter:
        if isinstance(arg, QueryFilter):
            return arg
        elif isinstance(arg, dict):
            return cls._from_dict(arg)
        else:
            raise ValueError(f"args must be a QueryFilter or a dict with 'name' and 'param' keys, got {arg}")

    @classmethod
    def _from_kwarg(cls, key: str, value: str) -> QueryFilter:
        if ":" in value:
            modifier_str, value = value.split(":", 1)
            modifier = Modifier(modifier_str)
        else:
            modifier = None
        return QueryFilter(name=key, param=value, modifier=modifier)


class Query(Generic[T], ABC):
    """Basic Query object with methods to manage filters
    Methods:
        filter: adds a filter to the query object, returns a new query object
        where: Applies a filter on the query object, returns a new query object
        collect: Executes the query on the client and returns the results
        _generic_execute_filter: Executes a filter on a list of objects
    """

    def __init__(
        self, client: Client, filters: Optional[List[QueryFilter]] = None, meta_params: Optional[Param] = None
    ):
        self.meta_params = meta_params
        self.client = client
        self.filters = filters
        self.results: Optional[List[T]] = None

    def filter(self, filter: QueryFilter) -> Query[T]:
        return self + filter

    def __add__(self, filter: QueryFilter) -> Query[T]:
        assert filter is not None
        assert isinstance(filter, QueryFilter)
        if self.filters is None:
            self.filters = []
        return self.__class__(self.client, filters=[*self.filters, filter], meta_params=self.meta_params)

    def __sub__(self, filter: QueryFilter) -> Query[T]:
        assert filter is not None
        assert isinstance(filter, QueryFilter)
        if self.filters is None:
            return self
        return self.__class__(
            self.client, filters=[f for f in self.filters if f != filter], meta_params=self.meta_params
        )

    def __iadd__(self, filter: QueryFilter) -> Query[T]:
        assert filter is not None
        assert isinstance(filter, QueryFilter)
        if self.filters is None:
            self.filters = [filter]
            return self
        self.filters.append(filter)
        return self

    def __isub__(self, filter: QueryFilter) -> Query[T]:
        assert filter is not None
        assert isinstance(filter, QueryFilter)
        if self.filters is None:
            return self
        self.filters = [f for f in self.filters if f != filter]
        return self

    def __len__(self) -> int:
        if self.results is None:
            self.results = list(self.collect())
        return len(self.results)

    def __iter__(self) -> Query[T]:
        self.n = 0
        return self

    def __next__(self) -> T:
        if self.results is None:
            self.results = list(self.collect())
        if self.n < len(self.results):
            result = self.results[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def __getitem__(self, index: int) -> T:
        if self.results is None:
            self.results = list(self.collect())
        return self.results[index]

    def __setitem__(self, index: int, value: T) -> None:
        if self.results is None:
            self.results = list(self.collect())
        self.results[index] = value

    def where(self, *args: object, **kwargs: str) -> Query[T]:
        filters = QueryFilter._from_args(*args, **kwargs)
        for item in filters:
            self += item
        return self

    @abstractmethod
    def collect(self) -> List[T]:
        raise NotImplementedError("Not implemented")

    def _generic_execute_filter(self, objects: List[T], filter: QueryFilter) -> List[T]:
        return [m for m in objects if filter.filter_attr(getattr(m._item, filter.name))]



from pathlib import Path

import pytest

from darwin.future.data_objects.workflow import WFStageConfig

test_data_path: Path = Path(__file__).parent / "data"
validate_json = test_data_path / "stage_config.json"


def test_file_exists() -> None:
    # This is a test sanity check to make sure the file exists
    # Helps avoids headaches when debugging tests
    assert validate_json.exists()


def test_WFStageConfig_validates_from_valid_json() -> None:
    parsed_stage_config = WFStageConfig.parse_file(validate_json)

    assert isinstance(parsed_stage_config, WFStageConfig)
